import unittest
import asyncio
import os
import shutil
import contextlib
import uuid
from datetime import datetime, timedelta
from stent import Stent, Result, RetryPolicy
from stent.executor import UnregisteredFunctionError, _original_sleep
from stent.registry import registry, FunctionRegistry, FunctionMetadata
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

# Define some test functions globally so pickle/registry can find them
@Stent.durable()
async def simple_task(x: int) -> int:
    return x * 2

@Stent.durable()
async def failing_task():
    raise ValueError("I failed")

@Stent.durable(retry_policy=RetryPolicy(max_attempts=3, initial_delay=0.01, backoff_factor=1.0))
async def retryable_task(succeed_on_attempt: int):
    pass

ATTEMPT_COUNTER = {}
RECOVERY_TEST_STATE = {"first_run": True}
DELAYED_RETRY_COUNTER = {}
FALSEY_IDEMPOTENT_COUNTER = {}

@Stent.durable()
async def recovery_task():
    # If it's the first run (simulated crash), we sleep to allow cancellation
    if RECOVERY_TEST_STATE["first_run"]:
        RECOVERY_TEST_STATE["first_run"] = False
        await _original_sleep(10)  # real sleep to simulate slow work for crash test
    return "recovered"

@Stent.durable(retry_policy=RetryPolicy(max_attempts=4, initial_delay=0.01))
async def stateful_retry_task(exec_id_for_counter: str):
    count = ATTEMPT_COUNTER.get(exec_id_for_counter, 0) + 1
    ATTEMPT_COUNTER[exec_id_for_counter] = count
    if count < 3:
        raise ValueError(f"Fail attempt {count}")
    return count


@Stent.durable(retry_policy=RetryPolicy(max_attempts=2, initial_delay=1.0, backoff_factor=1.0))
async def delayed_retry_task(case_id: str):
    count = DELAYED_RETRY_COUNTER.get(case_id, 0) + 1
    DELAYED_RETRY_COUNTER[case_id] = count
    if count == 1:
        raise ValueError("fail once")
    return count


@Stent.durable(idempotent=True)
async def falsey_idempotent_task(case_id: str, value):
    FALSEY_IDEMPOTENT_COUNTER[case_id] = FALSEY_IDEMPOTENT_COUNTER.get(case_id, 0) + 1
    return value


@Stent.durable()
async def falsey_idempotent_workflow(case_id: str, value):
    return await falsey_idempotent_task(case_id, value)

@Stent.durable(queue="high_priority_queue", tags=["data_processing"])
async def high_priority_data_task(data: str) -> str:
    return f"Processed {data} with high priority"

@Stent.durable(queue="low_priority_queue", tags=["reporting"])
async def low_priority_report_task(report_id: str) -> str:
    return f"Generated report {report_id}"


@Stent.durable(tags=["alpha"])
async def alpha_tagged_task(value: str) -> str:
    return f"alpha:{value}"


@Stent.durable(tags=["beta"])
async def beta_tagged_task(value: str) -> str:
    return f"beta:{value}"


@Stent.durable()
async def untagged_task(value: str) -> str:
    return f"untagged:{value}"

LONG_TASK_INVOCATIONS = 0

@Stent.durable()
async def guarded_long_activity(duration: float) -> int:
    global LONG_TASK_INVOCATIONS
    LONG_TASK_INVOCATIONS += 1
    await asyncio.sleep(duration)
    return LONG_TASK_INVOCATIONS


async def registry_isolated_workflow(value: int) -> int:
    return value + 5

DLQ_REPLAY_ATTEMPTS: dict[str, int] = {}

@Stent.durable(retry_policy=RetryPolicy(max_attempts=1, initial_delay=0.01))
async def flaky_once_task(key: str) -> str:
    attempt = DLQ_REPLAY_ATTEMPTS.get(key, 0)
    DLQ_REPLAY_ATTEMPTS[key] = attempt + 1
    if attempt == 0:
        raise RuntimeError("boom")
    return f"ok-{attempt}"


class TestExecution(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"{os.getpid()}_{id(self)}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)
            
    async def test_simple_execution(self):
        exec_id = await self.executor.dispatch(simple_task, 21)
        result = await self._wait_for_result(exec_id)
        self.assertEqual(result.value, 42)

    async def test_dispatch_requires_registered_function(self):
        async def local_function():
            return "nope"

        with self.assertRaises(UnregisteredFunctionError):
            await self.executor.dispatch(local_function)

    async def test_executor_can_use_custom_registry(self):
        with self.assertRaises(UnregisteredFunctionError):
            await self.executor.dispatch(registry_isolated_workflow, 2)

        custom_registry = FunctionRegistry()
        custom_registry.register(
            FunctionMetadata(
                name=custom_registry.name_for_function(registry_isolated_workflow),
                fn=registry_isolated_workflow,
                cached=False,
                retry_policy=RetryPolicy(),
                tags=[],
                priority=0,
                queue=None,
                idempotent=False,
                idempotency_key_func=None,
                version=None,
            )
        )

        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass

        custom_executor = Stent(backend=self.backend, function_registry=custom_registry)
        custom_worker = asyncio.create_task(custom_executor.serve(poll_interval=0.05))

        try:
            exec_id = await custom_executor.dispatch(registry_isolated_workflow, 5)
            result = await custom_executor.wait_for(exec_id, expiry=5.0)
            self.assertEqual(result.value, 10)
        finally:
            custom_worker.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await custom_worker
            await custom_executor.shutdown()
        
    async def test_failure_execution(self):
        exec_id = await self.executor.dispatch(failing_task)
        
        # Wait for completion
        while True:
            state = await self.executor.state_of(exec_id)
            if state.state in ("completed", "failed", "timed_out"):
                break
            await asyncio.sleep(0.1)
            
        state = await self.executor.state_of(exec_id)
        self.assertEqual(state.state, "failed")
        self.assertIn("I failed", str(state.result) if state.result else str(state))
        
    async def test_retry_logic(self):
        eid = "retry_test_1"
        ATTEMPT_COUNTER[eid] = 0
        
        exec_id = await self.executor.dispatch(stateful_retry_task, eid)
        
        result = await self._wait_for_result(exec_id)
        
        self.assertEqual(result.value, 3)
        self.assertEqual(ATTEMPT_COUNTER[eid], 3)
        
        tasks = await self.backend.list_tasks_for_execution(exec_id)
        root_task = next(t for t in tasks if t.kind == "orchestrator")
        self.assertEqual(root_task.retries, 2)

    async def test_retry_sets_scheduled_for_and_not_lease_expiry(self):
        case_id = str(uuid.uuid4())
        DELAYED_RETRY_COUNTER[case_id] = 0

        exec_id = await self.executor.dispatch(delayed_retry_task, case_id)

        pending_retry_seen = False
        for _ in range(80):
            tasks = await self.backend.list_tasks_for_execution(exec_id)
            root_task = next((t for t in tasks if t.kind == "orchestrator"), None)
            if root_task and root_task.retries == 1 and root_task.state == "pending":
                pending_retry_seen = True
                scheduled_for = root_task.scheduled_for
                self.assertIsNotNone(scheduled_for)
                self.assertIsNone(root_task.lease_expires_at)
                if scheduled_for is None:
                    self.fail("scheduled_for should be set for retry delay")
                self.assertGreater(scheduled_for.timestamp(), datetime.now().timestamp())
                await asyncio.sleep(0.15)
                root_task_after = await self.backend.get_task(root_task.id)
                self.assertIsNotNone(root_task_after)
                if root_task_after is None:
                    self.fail("Task unexpectedly missing")
                self.assertEqual(root_task_after.state, "pending")
                break
            await asyncio.sleep(0.05)

        self.assertTrue(pending_retry_seen, "Did not observe task in delayed retry pending state")

        result = await self.executor.wait_for(exec_id, expiry=10.0)
        self.assertEqual(result.or_raise(), 2)

    async def test_idempotency_reuses_falsey_results(self):
        cases = [
            ("zero", 0),
            ("false", False),
            ("empty-str", ""),
            ("empty-list", []),
            ("empty-dict", {}),
        ]

        for case_id, value in cases:
            FALSEY_IDEMPOTENT_COUNTER[case_id] = 0

            first_exec = await self.executor.dispatch(falsey_idempotent_workflow, case_id, value)
            first_result = await self.executor.wait_for(first_exec, expiry=5.0)
            self.assertEqual(first_result.or_raise(), value)

            second_exec = await self.executor.dispatch(falsey_idempotent_workflow, case_id, value)
            second_result = await self.executor.wait_for(second_exec, expiry=5.0)
            self.assertEqual(second_result.or_raise(), value)

            self.assertEqual(FALSEY_IDEMPOTENT_COUNTER[case_id], 1)

    async def test_replay_dead_letter(self):
        key = str(uuid.uuid4())
        exec_id = await self.executor.dispatch(flaky_once_task, key)

        while True:
            state = await self.executor.state_of(exec_id)
            if state.state in ("failed", "completed"):
                break
            await asyncio.sleep(0.05)

        self.assertEqual(state.state, "failed")
        letters = await self.executor.list_dead_letters()
        self.assertEqual(len(letters), 1)
        replayed_id = await self.executor.replay_dead_letter(letters[0].id)
        self.assertNotEqual(replayed_id, letters[0].id)
        result = await self.executor.wait_for(exec_id, expiry=5.0)
        self.assertTrue(result.ok)
        value = result.or_raise()
        self.assertIsInstance(value, str)
        self.assertIn("ok-", value)
        self.assertEqual(await self.executor.list_dead_letters(), [])

    async def test_queue_and_tags_filtering(self):
        # Stop default worker FIRST to prevent it from stealing tasks
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
            
        # Ensure any background tasks from default worker are done (though none should be there)
        await self.executor.shutdown()

        # Let's create tasks for different queues
        hp_exec_id = await self.executor.dispatch(high_priority_data_task, "important_data")
        lp_exec_id = await self.executor.dispatch(low_priority_report_task, "monthly_report")

        # Start a worker only for high_priority_queue
        hp_executor = Stent(backend=self.backend)
        hp_worker_task = asyncio.create_task(hp_executor.serve(queues=["high_priority_queue"], poll_interval=0.1))
        
        hp_result = await self._wait_for_result(hp_exec_id)
        self.assertEqual(hp_result.value, "Processed important_data with high priority")
        
        # Verify low priority task is still pending
        lp_state = await self.executor.state_of(lp_exec_id)
        self.assertEqual(lp_state.state, "pending")

        hp_worker_task.cancel()
        try:
            await hp_worker_task
        except asyncio.CancelledError:
            pass
        await hp_executor.shutdown()

        # Start a worker for low_priority_queue
        lp_executor = Stent(backend=self.backend)
        lp_worker_task = asyncio.create_task(lp_executor.serve(queues=["low_priority_queue"], poll_interval=0.1))
        
        lp_result = await self._wait_for_result(lp_exec_id)
        self.assertEqual(lp_result.value, "Generated report monthly_report")

        lp_worker_task.cancel()
        try:
            await lp_worker_task
        except asyncio.CancelledError:
            pass
        await lp_executor.shutdown()

    async def test_worker_tags_filtering(self):
        # Stop default worker so only explicitly configured workers consume tasks.
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()

        alpha_exec_id = await self.executor.dispatch(alpha_tagged_task, "A")
        beta_exec_id = await self.executor.dispatch(beta_tagged_task, "B")
        untagged_exec_id = await self.executor.dispatch(untagged_task, "U")

        alpha_executor = Stent(backend=self.backend)
        alpha_worker_task = asyncio.create_task(alpha_executor.serve(tags=["alpha"], poll_interval=0.1))

        try:
            alpha_result = await self._wait_for_result(alpha_exec_id)
            self.assertEqual(alpha_result.value, "alpha:A")

            untagged_result = await self._wait_for_result(untagged_exec_id)
            self.assertEqual(untagged_result.value, "untagged:U")

            beta_state = await self.executor.state_of(beta_exec_id)
            self.assertEqual(beta_state.state, "pending")
        finally:
            alpha_worker_task.cancel()
            try:
                await alpha_worker_task
            except asyncio.CancelledError:
                pass
            await alpha_executor.shutdown()

        beta_executor = Stent(backend=self.backend)
        beta_worker_task = asyncio.create_task(beta_executor.serve(tags=["beta"], poll_interval=0.1))

        try:
            beta_result = await self._wait_for_result(beta_exec_id)
            self.assertEqual(beta_result.value, "beta:B")
        finally:
            beta_worker_task.cancel()
            try:
                await beta_worker_task
            except asyncio.CancelledError:
                pass
            await beta_executor.shutdown()

    async def test_lease_expiration_crash(self):
        RECOVERY_TEST_STATE["first_run"] = True
        
        # 1. Stop default worker
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
            
        # 2. Start worker with short lease
        short_lease = timedelta(seconds=1)
        worker1 = asyncio.create_task(self.executor.serve(lease_duration=short_lease, poll_interval=0.1))
        
        # 3. Dispatch task
        exec_id = await self.executor.dispatch(recovery_task)
        
        # 4. Wait for it to be running
        target_task: asyncio.Task | None = None
        start_wait = asyncio.get_running_loop().time()
        while asyncio.get_running_loop().time() - start_wait < 5:
            tasks = await self.backend.list_tasks_for_execution(exec_id)
            root_task = next((t for t in tasks if t.kind == "orchestrator"), None)
            
            if root_task and root_task.state == "running":
                # Try to find the python task corresponding to this
                current = asyncio.current_task()
                for t in asyncio.all_tasks():
                    if t is current or t is worker1:
                        continue
                    # Check for _handle_task in coroutine name
                    if "handle_task" in str(t) or "handle_task" in repr(t):
                        target_task = t
                        break
                if target_task:
                    break
            await asyncio.sleep(0.05)
            
        self.assertIsNotNone(target_task, "Could not find worker handler task")
        
        # 5. Simulate Crash: Cancel the handler task
        if target_task:
            target_task.cancel()
            try:
                await target_task
            except asyncio.CancelledError:
                pass
            
        # Stop worker1 loop too
        worker1.cancel()
        try:
            await worker1
        except asyncio.CancelledError:
            pass
            
        # 6. Verify state is still "running" (simulating crash before update)
        tasks = await self.backend.list_tasks_for_execution(exec_id)
        root_task = next(t for t in tasks if t.kind == "orchestrator")
        self.assertEqual(root_task.state, "running")
        
        # 7. Wait for lease to expire
        await asyncio.sleep(1.5) 
        
        # 8. Start worker2
        worker2 = asyncio.create_task(self.executor.serve(poll_interval=0.1))
        
        # 9. Wait for result
        result = await self._wait_for_result(exec_id)
        self.assertEqual(result.value, "recovered")
        
        worker2.cancel()
        try:
            await worker2
        except asyncio.CancelledError:
            pass

    async def test_lease_renewal_prevents_duplicate_execution(self):
        global LONG_TASK_INVOCATIONS
        LONG_TASK_INVOCATIONS = 0

        # Stop default worker to configure heartbeat workers
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass

        lease_duration = timedelta(seconds=0.3)
        heartbeat_interval = timedelta(seconds=0.1)

        worker_exec1 = Stent(backend=self.backend)
        worker_exec2 = Stent(backend=self.backend)
        worker1 = asyncio.create_task(
            worker_exec1.serve(
                lease_duration=lease_duration,
                heartbeat_interval=heartbeat_interval,
                poll_interval=0.05,
            )
        )
        worker2 = asyncio.create_task(
            worker_exec2.serve(
                lease_duration=lease_duration,
                heartbeat_interval=heartbeat_interval,
                poll_interval=0.05,
            )
        )

        try:
            exec_id = await self.executor.dispatch(guarded_long_activity, 0.8)
            result = await self._wait_for_result(exec_id)
            self.assertEqual(result.value, 1)
            self.assertEqual(LONG_TASK_INVOCATIONS, 1)
        finally:
            worker1.cancel()
            worker2.cancel()
            for t in (worker1, worker2):
                try:
                    await t
                except asyncio.CancelledError:
                    pass
            await worker_exec1.shutdown()
            await worker_exec2.shutdown()

    async def test_cleanup(self):
        # Stop default worker to manually control execution
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
            
        # 1. Run a task to completion
        # We need a worker for this
        worker = asyncio.create_task(self.executor.serve(poll_interval=0.1))
        
        exec_id = await self.executor.dispatch(simple_task, 99)
        result = await self._wait_for_result(exec_id)
        self.assertEqual(result.value, 198)
        
        # 2. Dispatch a task that will stay pending (we'll stop worker)
        worker.cancel()
        try:
            await worker
        except asyncio.CancelledError:
            pass
            
        exec_pending_id = await self.executor.dispatch(simple_task, 100)
        
        # 3. Cleanup with future cutoff
        # This should delete the completed task, but NOT the pending one
        cutoff = datetime.now() + timedelta(days=1)
        
        count = await self.backend.cleanup_executions(cutoff)
        self.assertGreaterEqual(count, 1)
        
        # Verify completed is gone
        rec = await self.backend.get_execution(exec_id)
        self.assertIsNone(rec)
        
        # Verify pending is present
        rec_pending = await self.backend.get_execution(exec_pending_id)
        self.assertIsNotNone(rec_pending)
        if rec_pending is None:
            self.fail("Pending execution should still exist")
        self.assertEqual(rec_pending.state, "pending")


    async def _wait_for_result(self, exec_id):
        while True:
            state = await self.executor.state_of(exec_id)
            if state.state in ("completed", "failed", "timed_out"):
                break
            await asyncio.sleep(0.1)
        return await self.executor.result_of(exec_id)


class TestAutoDispatch(unittest.IsolatedAsyncioTestCase):
    """Tests for the Stent.use() / auto-dispatch feature."""

    async def asyncSetUp(self):
        self.backend = get_test_backend(f"autodispatch_{os.getpid()}_{id(self)}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)

    async def asyncTearDown(self):
        Stent.reset()
        await cleanup_test_backend(self.backend)

    def test_init_stores_default_executor(self):
        self.assertIsNone(Stent._default_executor)
        executor = Stent.use(backend=self.backend)
        self.assertIs(Stent._default_executor, executor)

    def test_reset_clears_default_executor(self):
        Stent.use(backend=self.backend)
        self.assertIsNotNone(Stent._default_executor)
        Stent.reset()
        self.assertIsNone(Stent._default_executor)

    async def test_auto_dispatch_returns_result(self):
        """Calling a durable function with a default executor auto-dispatches and returns the result."""
        executor = Stent.use(backend=self.backend)
        worker = asyncio.create_task(executor.serve(poll_interval=0.05))
        try:
            result = await simple_task(21)
            self.assertEqual(result, 42)
        finally:
            worker.cancel()
            try:
                await worker
            except asyncio.CancelledError:
                pass
            await executor.shutdown()

    async def test_auto_dispatch_propagates_errors(self):
        """Errors from auto-dispatched functions are re-raised to the caller."""
        executor = Stent.use(backend=self.backend)
        worker = asyncio.create_task(executor.serve(poll_interval=0.05))
        try:
            with self.assertRaises(Exception):
                await failing_task()
        finally:
            worker.cancel()
            try:
                await worker
            except asyncio.CancelledError:
                pass
            await executor.shutdown()

    async def test_no_default_executor_runs_locally(self):
        """Without a default executor, durable functions still run locally."""
        self.assertIsNone(Stent._default_executor)
        result = await simple_task(10)
        self.assertEqual(result, 20)



from __future__ import annotations

import asyncio
from datetime import datetime
from typing import Awaitable, Callable

from stent.core import ExecutionState, TaskRecord
from stent.notifications.base import NotificationBackend

TASK_TERMINAL_STATES = ("completed", "failed")
EXECUTION_TERMINAL_STATES = ("completed", "failed", "timed_out", "cancelled")


async def wait_for_task_terminal(
    *,
    task_id: str,
    get_task: Callable[[str], Awaitable[TaskRecord | None]],
    notification_backend: NotificationBackend | None,
    expiry: float | None,
    poll_min_interval: float,
    poll_max_interval: float,
    poll_backoff_factor: float,
    sleep_fn: Callable[[float], Awaitable[None]],
    expiry_error_factory: Callable[[str], Exception],
) -> TaskRecord:
    if notification_backend:
        it = notification_backend.subscribe_to_task(task_id, expiry=expiry)
        try:
            async for _ in it:
                pass
        except asyncio.TimeoutError as exc:
            raise expiry_error_factory(f"Task {task_id} timed out") from exc

        task = await get_task(task_id)
        if not task:
            raise ValueError(f"Task not found after notification: {task_id}")
        if task.state not in TASK_TERMINAL_STATES:
            if expiry is not None:
                raise expiry_error_factory(f"Task {task_id} timed out")
            raise Exception(f"Task {task_id} is not terminal after notification")
        return task

    start = datetime.now()
    delay = poll_min_interval
    while True:
        task = await get_task(task_id)
        if task and task.state in TASK_TERMINAL_STATES:
            return task
        if expiry and (datetime.now() - start).total_seconds() > expiry:
            raise expiry_error_factory(f"Task {task_id} timed out")
        await sleep_fn(delay)
        delay = min(poll_max_interval, delay * poll_backoff_factor)


async def wait_for_execution_terminal(
    *,
    execution_id: str,
    state_of: Callable[[str], Awaitable[ExecutionState]],
    notification_backend: NotificationBackend | None,
    expiry: float | None,
    sleep_interval: float,
    expiry_error_factory: Callable[[str], Exception],
) -> ExecutionState:
    if notification_backend:
        it = notification_backend.subscribe_to_execution(execution_id, expiry=expiry)
        try:
            async for _ in it:
                pass
        except asyncio.TimeoutError:
            raise expiry_error_factory(f"Timed out waiting for execution {execution_id}")

        state = await state_of(execution_id)
        if state.state not in EXECUTION_TERMINAL_STATES:
            if expiry is not None:
                raise expiry_error_factory(f"Timed out waiting for execution {execution_id}")
            raise Exception(f"Execution {execution_id} is not terminal after notification")
        return state

    start = datetime.now()
    while True:
        state = await state_of(execution_id)
        if state.state in EXECUTION_TERMINAL_STATES:
            return state

        if expiry and (datetime.now() - start).total_seconds() > expiry:
            raise expiry_error_factory(f"Timed out waiting for execution {execution_id}")

        await asyncio.sleep(sleep_interval)



import unittest
import asyncio
import uuid
import os
from datetime import datetime, timedelta
from stent import Stent
from stent.registry import registry
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

# Globals for pickles
@Stent.durable(max_concurrent=1)
async def limited_task(idx: int):
    # We log to a global list (careful with parallel tests, but IsolatedAsyncioTestCase runs sequentially usually? No, but global var is shared)
    # Better to return timestamps
    await Stent.sleep(timedelta(seconds=0.5))
    return (idx, datetime.now())

@Stent.durable()
async def orchestrator():
    t1 = limited_task(1)
    t2 = limited_task(2)
    t3 = limited_task(3)
    return await Stent.gather(t1, t2, t3)

@Stent.durable(max_concurrent=1)
async def serial_task(idx: int):
    await Stent.sleep(timedelta(seconds=0.2))
    return ("serial", idx, datetime.now())

@Stent.durable(max_concurrent=5)
async def parallel_task(idx: int):
    await Stent.sleep(timedelta(seconds=0.2))
    return ("parallel", idx, datetime.now())

@Stent.durable()
async def mixed_workflow():
    return await Stent.gather(
        serial_task(1), serial_task(2),
        parallel_task(1), parallel_task(2)
    )

class TestRateLimit(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.test_id = str(uuid.uuid4())
        self.backend = get_test_backend(self.test_id)
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1, max_concurrency=10))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_max_concurrent_limit(self):
        exec_id = await self.executor.schedule(timedelta(seconds=0), orchestrator)
        result_wrapper = await self.executor.wait_for(exec_id, expiry=30.0) # Increased expiry
        result = result_wrapper.or_raise()
        
        # Result is list of (idx, timestamp)
        # Sort by idx just in case gather returns in order (it does usually)
        # But we want to check timestamps
        
        # Sort by completion time
        # Handle string timestamps from JSON serialization
        sorted_res = sorted(result, key=lambda x: datetime.fromisoformat(x[1]) if isinstance(x[1], str) else x[1])
        
        start_times = [datetime.fromisoformat(x[1]) if isinstance(x[1], str) else x[1] for x in sorted_res]
        
        # Check gaps. Since each task takes 0.5s and max_concurrent=1,
        # completion times should be at least 0.5s apart.
        for i in range(len(start_times) - 1):
            diff = (start_times[i+1] - start_times[i]).total_seconds()
            print(f"Diff {i}: {diff}")
            # Allow some slack, but should be close to 0.5
            # If parallel, diff would be small (e.g. 0.05)
            self.assertGreaterEqual(diff, 0.4, f"Tasks completed too close together: {diff}s. Expected sequential execution.")

    async def test_mixed_concurrency(self):
        exec_id = await self.executor.schedule(timedelta(0), mixed_workflow)
        result_wrapper = await self.executor.wait_for(exec_id, expiry=30.0) # Increased expiry
        result = result_wrapper.or_raise()
        
        # Convert timestamps
        def convert(item):
            ts = item[2]
            if isinstance(ts, str):
                ts = datetime.fromisoformat(ts)
            return [item[0], item[1], ts] # Use list as original might be list
            
        result = [convert(x) for x in result]
        
        serial_res = sorted([x for x in result if x[0] == "serial"], key=lambda x: x[2])
        parallel_res = sorted([x for x in result if x[0] == "parallel"], key=lambda x: x[2])
        
        self.assertEqual(len(serial_res), 2)
        self.assertEqual(len(parallel_res), 2)
        
        # Serial tasks: completion times separated by ~0.2s
        diff_serial = (serial_res[1][2] - serial_res[0][2]).total_seconds()
        print(f"Diff serial: {diff_serial}")
        self.assertGreaterEqual(diff_serial, 0.15)
        
        # Parallel tasks: completion times close (both start at same time, sleep 0.2s, finish at same time)
        diff_parallel = abs((parallel_res[1][2] - parallel_res[0][2]).total_seconds())
        print(f"Diff parallel: {diff_parallel}")
        # Relaxed constraint for remote DB
        self.assertLess(diff_parallel, 2.0)


import asyncio
import os
import random
import time
import logging
import traceback
import multiprocessing
from datetime import datetime, timedelta
from stent import Stent, Result, RetryPolicy

# Disable extensive logging for the dashboard effect, only errors
logging.basicConfig(level=logging.ERROR)

# --- Configuration ---
NUM_ACTIVITY_WORKERS = 5
NUM_ORCHESTRATOR_WORKERS = 2
DB_PATH = "media_pipeline.sqlite"
LEASE_DURATION = timedelta(seconds=5)

# --- Activities ---

ACTIVITY_QUEUE = "processing_queue"

@Stent.durable(queue=ACTIVITY_QUEUE)
async def validate_upload(file_name: str) -> bool:
    await asyncio.sleep(random.uniform(0.5, 1.5))
    if "corrupt" in file_name:
        raise ValueError(f"File {file_name} header is corrupt")
    return True

@Stent.durable(retry_policy=RetryPolicy(max_attempts=3, initial_delay=1.0), queue=ACTIVITY_QUEUE)
async def scan_for_safety(file_name: str) -> str:
    await asyncio.sleep(random.uniform(1.0, 2.5))
    if random.random() < 0.05: raise ConnectionError("Safety API expiry")
    if random.random() < 0.02: return "flagged"
    return "clean"

@Stent.durable(queue=ACTIVITY_QUEUE, cached=True)
async def extract_audio(file_name: str) -> str:
    await asyncio.sleep(random.uniform(1.5, 3.0))
    return f"{file_name}.wav"

@Stent.durable(queue="gpu_tasks", cached=True)
async def transcribe_audio(audio_file: str) -> str:
    duration = random.uniform(3.0, 6.0)
    await asyncio.sleep(duration)
    return f"Transcription({int(duration)}s)"

@Stent.durable(queue="gpu_tasks", cached=True)
async def generate_thumbnails(file_name: str) -> list[str]:
    await asyncio.sleep(random.uniform(2.0, 4.0))
    return [f"thumb_{i}.jpg" for i in range(3)]

@Stent.durable(queue=ACTIVITY_QUEUE, cached=True)
async def generate_metadata(transcription: str) -> dict:
    await asyncio.sleep(random.uniform(1.0, 2.0))
    return {"tags": ["viral"], "sentiment": "positive"}

@Stent.durable(queue=ACTIVITY_QUEUE)
async def package_assets(file_name: str, transcription: str, thumbs: list[str], metadata: dict) -> str:
    await asyncio.sleep(random.uniform(0.5, 1.0))
    return f"s3://{file_name}.zip"

# --- Orchestrator ---

@Stent.durable()
async def media_processing_pipeline(file_name: str) -> Result[dict, Exception]:
    try:
        await validate_upload(file_name)
        
        status = await scan_for_safety(file_name)
        if status == "flagged": return Result.Error(Exception("Content flagged"))
            
        # Parallel branches
        results = await asyncio.gather(
            process_audio_chain(file_name),
            generate_thumbnails(file_name)
        )
        (transcription, metadata), thumbs = results
        
        final_url = await package_assets(file_name, transcription, thumbs, metadata)
        return Result.Ok({"status": "success", "url": final_url})
    except Exception as e:
        return Result.Error(e)

@Stent.durable()
async def process_audio_chain(file_name: str):
    audio_path = await extract_audio(file_name)
    text = await transcribe_audio(audio_path)
    meta = await generate_metadata(text)
    return text, meta

# --- Worker Process Entry Point ---

def run_worker_process(db_path, queues, worker_id, lease_duration_seconds):
    # Re-init logging for process
    logging.basicConfig(level=logging.ERROR)
    
    async def _run():
        backend = Stent.backends.SQLiteBackend(db_path)
        # No init_db here, assumed done by main
        executor = Stent(backend=backend)
        await executor.serve(
            max_concurrency=1,
            poll_interval=0.1,
            queues=queues,
            worker_id=worker_id,
            lease_duration=timedelta(seconds=lease_duration_seconds)
        )

    try:
        asyncio.run(_run())
    except KeyboardInterrupt:
        pass

# --- Dashboard & Runner ---

def clear_screen():
    print("\033[H\033[J", end="")

def print_dashboard(statuses: list[dict], worker_activities: list, start_time: float, crash_count: int, total_jobs: int):
    clear_screen()
    elapsed = time.time() - start_time
    completed_count = sum(1 for s in statuses if s['state'] in ('completed', 'failed', 'timed_out'))
    
    print(f"--- Resilience Demo: Media Pipeline ({elapsed:.1f}s) ---")
    print(f"Active Worker Processes: {len(worker_activities)} | Total Process Crashes: {crash_count}")
    print(f"Total Jobs Dispatched: {total_jobs} | Completed: {completed_count} | Active: {len(statuses) - completed_count}")
    
    # --- Jobs Table (Last 15) ---
    print("\n[ LATEST JOBS ]")
    print("-" * 100)
    print(f"{'Job ID':<38} | {'State':<10} | {'Current Step':<35} | {'Steps'}")
    print("-" * 100)
    
    # Show recent jobs first
    for s in statuses[-20:]:
        progress_parts = s['progress'].split(" > ")
        last_step = progress_parts[-1] if progress_parts else "Initializing"
        
        cache_hits = sum(1 for p in progress_parts if "cache_hit" in p)
        if cache_hits > 0:
            last_step += f" ({cache_hits} recovered)"

        if len(last_step) > 35: last_step = "..." + last_step[-32:]
            
        color = ""
        if s['state'] == "running": color = "\033[94m" # Blue
        elif s['state'] == "completed": color = "\033[92m" # Green
        elif s['state'] == "failed": color = "\033[91m" # Red
        reset = "\033[0m"
        
        print(f"{s['id']:<38} | {color}{s['state']:<10}{reset} | {last_step:<35} | {len(progress_parts)}")
    
    # --- Workers Table ---
    print("\n[ WORKERS ]")
    print("-" * 100)
    print(f"{'Worker ID':<20} | {'Type':<12} | {'Task':<45} | {'Age'}")
    print("-" * 100)
    
    worker_activities.sort(key=lambda x: x.get('type', ''))
    
    for w in worker_activities:
        task = w.get('task')
        wid = w['id']
        wtype = w['type']
        
        if task:
            step = task.step_name
            if task.retries > 0:
                step += f" (Retry {task.retries})"
            
            duration_s = ""
            if task.started_at:
                d = (datetime.now() - task.started_at).total_seconds()
                duration_s = f"{d:.1f}s"
            
            print(f"{wid:<20} | {wtype:<12} | {step:<45} | {duration_s}")
        else:
            print(f"{wid:<20} | {wtype:<12} | {'Idle / Polling...':<45} | -")
    
    print("-" * 100)

async def main():
    # Use spawn for consistency across platforms
    multiprocessing.set_start_method("spawn", force=True)
    
    if os.path.exists(DB_PATH): os.remove(DB_PATH)
    backend = Stent.backends.SQLiteBackend(DB_PATH)
    await backend.init_db()
    executor = Stent(backend=backend)
    
    # We store metadata about processes here
    # List of dicts: {'id': str, 'type': str, 'process': Process, 'queues': list}
    active_processes = []
    
    def spawn_worker(wtype, queues, index):
        wid = f"{wtype}-{index}"
        p = multiprocessing.Process(
            target=run_worker_process,
            args=(DB_PATH, queues, wid, LEASE_DURATION.total_seconds()),
            name=wid
        )
        p.start()
        active_processes.append({
            'id': wid,
            'type': wtype,
            'process': p,
            'queues': queues,
            'index': index
        })

    def ensure_workers():
        # Check for dead processes and remove them from list (Chaos Monkey kills them)
        # Actually Chaos Monkey will kill them, we just need to respawn missing ones.
        
        # Count active by type/index
        # We want strict slots: Activity-0 to Activity-4, Orchestrator-0 to Orchestrator-1
        
        existing_map = {(w['type'], w['index']): w for w in active_processes if w['process'].is_alive()}
        
        # Update active list to only alive ones
        active_processes[:] = existing_map.values()
        
        for i in range(NUM_ACTIVITY_WORKERS):
            if ('Activity', i) not in existing_map:
                spawn_worker("Activity", [ACTIVITY_QUEUE, "gpu_tasks"], i)
                
        for i in range(NUM_ORCHESTRATOR_WORKERS):
            if ('Orchestrator', i) not in existing_map:
                spawn_worker("Orchestrator", [None], i)

    ensure_workers()
    
    job_ids = []
    crash_count = 0
    start_time = time.time()
    
    # --- Loops ---
    
    async def spawner():
        i = 0
        while True:
            fname = f"video_{{i:03d}}.mp4"
            if i % 10 == 0: fname = "corrupt.mp4"
            eid = await executor.dispatch(media_processing_pipeline, fname)
            job_ids.append(eid)
            i += 1
            await asyncio.sleep(5.0)

    async def chaos_monkey():
        nonlocal crash_count
        while True:
            await asyncio.sleep(random.uniform(5.0, 8.0))
            if active_processes:
                victim = random.choice(active_processes)
                if victim['process'].is_alive():
                    victim['process'].terminate()
                    victim['process'].join()
                    crash_count += 1
                    # Remove from list immediately so main loop sees it missing
                    if victim in active_processes:
                        active_processes.remove(victim)
                    ensure_workers() # Respawn immediately

    spawner_task = asyncio.create_task(spawner())
    chaos_task = asyncio.create_task(chaos_monkey())
    
    try:
        while True:
            ensure_workers()
            
            # Prune job history
            if len(job_ids) > 50:
                job_ids[:] = job_ids[-50:]
            
            statuses = []
            for eid in job_ids:
                state = await executor.state_of(eid)
                statuses.append({
                    "id": eid,
                    "state": state.state,
                    "progress": state.progress_str
                })
            
            # Get DB view of who is working
            running_tasks = await executor.get_running_activities()
            worker_map = {t.worker_id: t for t in running_tasks}
            
            # Prepare display data merging Process list with DB tasks
            display_workers = []
            for w in active_processes:
                w_copy = {'id': w['id'], 'type': w['type']}
                w_copy['task'] = worker_map.get(w['id'])
                display_workers.append(w_copy)
            
            print_dashboard(statuses, display_workers, start_time, crash_count, len(job_ids))
            
            await asyncio.sleep(0.2)
            
    except asyncio.CancelledError:
        pass
    finally:
        spawner_task.cancel()
        chaos_task.cancel()
        for w in active_processes:
            if w['process'].is_alive():
                w['process'].terminate()
        print("\nSimulation Stopped.")

if __name__ == "__main__":
    asyncio.run(main())


from typing import Protocol, List, Optional
from datetime import datetime, timedelta
from stent.core import ExecutionRecord, TaskRecord, ExecutionProgress, SignalRecord, DeadLetterRecord

class Backend(Protocol):
    async def init_db(self) -> None: ...
    async def close(self) -> None: ...
    # executions
    async def create_execution(self, record: ExecutionRecord) -> None: ...
    async def create_execution_with_root_task(self, record: ExecutionRecord, task: TaskRecord) -> None: ...
    async def get_execution(self, execution_id: str) -> ExecutionRecord | None: ...
    async def update_execution(self, record: ExecutionRecord) -> None: ...
    async def list_executions(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[ExecutionRecord]: ...
    async def count_executions(self, state: str | None = None) -> int: ...

    # tasks
    async def create_task(self, task: TaskRecord) -> None: ...
    async def create_tasks(self, tasks: List[TaskRecord]) -> None: ...
    async def get_task(self, task_id: str) -> TaskRecord | None: ...
    async def update_task(self, task: TaskRecord) -> None: ...
    async def list_tasks(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[TaskRecord]: ...
    async def count_tasks(self, queue: str | None = None, state: str | None = None) -> int: ...

    async def claim_next_task(
        self,
        *,
        worker_id: str,
        queues: List[str] | None = None,
        tags: List[str] | None = None,
        now: datetime | None = None,
        lease_duration: timedelta | None = None,
        concurrency_limits: dict[str, int] | None = None,
    ) -> TaskRecord | None: ...

    async def renew_task_lease(
        self,
        task_id: str,
        worker_id: str,
        lease_duration: timedelta,
    ) -> bool: ...

    async def list_tasks_for_execution(self, execution_id: str) -> List[TaskRecord]: ...

    # progress
    async def append_progress(
        self,
        execution_id: str,
        progress: ExecutionProgress,
    ) -> None: ...

    # cache
    async def get_cached_result(self, cache_key: str) -> bytes | None: ...
    async def set_cached_result(self, cache_key: str, value: bytes, ttl: timedelta | None = None) -> None: ...

    async def get_idempotency_result(self, idempotency_key: str) -> bytes | None: ...
    async def set_idempotency_result(self, idempotency_key: str, value: bytes) -> None: ...

    # dead-letter
    async def move_task_to_dead_letter(
        self,
        task: TaskRecord,
        reason: str,
    ) -> None: ...
    async def list_dead_tasks(self, limit: int = 50) -> List[DeadLetterRecord]: ...
    async def count_dead_tasks(self) -> int: ...
    async def get_dead_task(self, task_id: str) -> DeadLetterRecord | None: ...
    async def delete_dead_task(self, task_id: str) -> bool: ...

    async def cleanup_executions(self, older_than: datetime) -> int: ...
    async def cleanup_dead_letters(self, older_than: datetime) -> int: ...

    # signals
    async def create_signal(self, signal: SignalRecord) -> None: ...
    async def get_signal(self, execution_id: str, name: str) -> SignalRecord | None: ...

    # execution context (counters/custom state)
    async def ensure_execution_counters(self, execution_id: str, counters: dict[str, int | float]) -> None: ...
    async def increment_execution_counter(self, execution_id: str, name: str, amount: int | float) -> float: ...
    async def get_execution_counters(self, execution_id: str) -> dict[str, int | float]: ...

    async def ensure_execution_state_values(self, execution_id: str, values: dict[str, bytes]) -> None: ...
    async def set_execution_state_value(self, execution_id: str, key: str, value: bytes) -> None: ...
    async def get_execution_state_values(self, execution_id: str) -> dict[str, bytes]: ...



import asyncio
import unittest
import os
from datetime import timedelta
from stent import Stent, Result
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

@Stent.durable()
async def long_running_task(duration_sec: float) -> str:
    await asyncio.sleep(duration_sec)
    return "done"

class TestMaxDuration(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"max_duration_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_max_duration_success(self):
        # Task takes 0.1s, max_duration 1s -> should pass
        # Increase max_duration to 5s to account for latency
        exec_id = await self.executor.dispatch(long_running_task, 0.1, max_duration="5s")
        # Increase wait_for expiry too
        result = await self.executor.wait_for(exec_id, expiry=10.0)
        self.assertEqual(result.value, "done")

    async def test_max_duration_timeout(self):
        # Task takes 2s, max_duration 1s -> should timeout
        # Using a very short max_duration to speed up test
        exec_id = await self.executor.dispatch(long_running_task, 2.0, max_duration=timedelta(seconds=0.5))
        
        # Wait for completion (it should fail/timeout)
        try:
            await self.executor.wait_for(exec_id)
        except Exception:
            pass # Expect error result, but wait_for might raise or return Result.Error depending on impl
            
        state = await self.executor.state_of(exec_id)
        self.assertIn(state.state, ("timed_out", "failed"))
        if state.state == "failed":
             # If it failed due to timeout exception bubbling up
             try:
                 await self.executor.result_of(exec_id)
             except Exception as e:
                 self.assertIn("timed out", str(e))

    async def test_conflict_error(self):
        with self.assertRaises(ValueError):
            await self.executor.dispatch(long_running_task, 0.1, expiry="1m", max_duration="1m")




import unittest
import asyncio
import os
import logging
from stent import Stent, Result, RetryPolicy
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@Stent.durable()
async def square(x: int) -> int:
    await asyncio.sleep(0.1)
    return x * x

@Stent.durable(retry_policy=RetryPolicy(max_attempts=3, initial_delay=0.1))
async def fail_on_three(x: int) -> int:
    if x == 3:
        raise ValueError("Three is forbidden")
    return x * x

@Stent.durable()
async def add(a: int, b: int) -> int:
    return a + b

class TestMap(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"map_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1, max_concurrency=10))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_map_basic(self):
        @Stent.durable()
        async def workflow(items: list[int]) -> list[int]:
            return await Stent.map(square, items)
    
        exec_id = await self.executor.dispatch(workflow, [1, 2, 3, 4, 5])
    
        # Wait for result
        final = await self.executor.wait_for(exec_id, expiry=30.0)
        self.assertTrue(final.ok)
        self.assertEqual(final.value, [1, 4, 9, 16, 25])

    async def test_map_empty(self):
        @Stent.durable()
        async def workflow() -> list[int]:
            return await Stent.map(square, [])

        exec_id = await self.executor.dispatch(workflow)
        final = await self.executor.wait_for(exec_id, expiry=5.0)
        self.assertTrue(final.ok)
        self.assertEqual(final.value, [])

    async def test_map_failure(self):
        @Stent.durable(retry_policy=RetryPolicy(max_attempts=1))
        async def workflow(items: list[int]) -> list[int]:
            return await Stent.map(fail_on_three, items)

        exec_id = await self.executor.dispatch(workflow, [1, 2, 3])
        
        final = await self.executor.wait_for(exec_id, expiry=15.0)
        self.assertFalse(final.ok)
        self.assertIn("Three is forbidden", str(final.error))

    async def test_gather_alias(self):
         from typing import Any
         @Stent.durable()
         async def workflow() -> list[Any]:
             t1 = square(10)
             t2 = square(20)
             return await Stent.gather(t1, t2)

         exec_id = await self.executor.dispatch(workflow)
         final = await self.executor.wait_for(exec_id, expiry=15.0)
         self.assertTrue(final.ok)
         self.assertEqual(final.value, [100, 400])

    async def test_dot_map(self):
        """square.map([1, 2, 3]) dispatches each item as a single-arg call."""
        @Stent.durable()
        async def workflow(items: list[int]) -> list[int]:
            return await square.map(items)

        exec_id = await self.executor.dispatch(workflow, [1, 2, 3, 4, 5])
        final = await self.executor.wait_for(exec_id, expiry=30.0)
        self.assertTrue(final.ok)
        self.assertEqual(final.value, [1, 4, 9, 16, 25])

    async def test_dot_map_empty(self):
        @Stent.durable()
        async def workflow() -> list[int]:
            return await square.map([])

        exec_id = await self.executor.dispatch(workflow)
        final = await self.executor.wait_for(exec_id, expiry=5.0)
        self.assertTrue(final.ok)
        self.assertEqual(final.value, [])

    async def test_starmap(self):
        """add.starmap([(1, 2), (3, 4)]) unpacks each tuple as positional args."""
        @Stent.durable()
        async def workflow() -> list[int]:
            return await add.starmap([(1, 2), (3, 4), (10, 20)])

        exec_id = await self.executor.dispatch(workflow)
        final = await self.executor.wait_for(exec_id, expiry=15.0)
        self.assertTrue(final.ok)
        self.assertEqual(final.value, [3, 7, 30])

    async def test_dot_map_local_fallback(self):
        """Without executor context, .map() falls back to local gather."""
        result = await square.map([2, 3, 4])
        self.assertEqual(result, [4, 9, 16])



"""
Stent CLI - Command-line interface for managing Stent executions.

Usage:
    stent [--db PATH] <command> [options]

Commands:
    list        List executions
    show        Show execution details
    tasks       List/show tasks
    stats       Show queue statistics
    dlq         Dead-letter queue operations
    cleanup     Clean up old executions
    signal      Send signal to execution
    watch       Live monitoring (requires rich)

Environment:
    STENT_DB  Default database path (default: stent.sqlite)
"""
import argparse
import asyncio
import json
import os
import sys
from datetime import datetime, timedelta
from typing import Any, cast

from stent import Stent, ExecutionState
from stent.core import TaskRecord, DeadLetterRecord

# ANSI color codes (fallback when rich is not available)
class Colors:
    RESET = "\033[0m"
    BOLD = "\033[1m"
    DIM = "\033[2m"
    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"

    @classmethod
    def disable(cls):
        for attr in ["RESET", "BOLD", "DIM", "RED", "GREEN", "YELLOW", "BLUE", "MAGENTA", "CYAN"]:
            setattr(cls, attr, "")


# Disable colors if not a TTY or on Windows without ANSI support
if not sys.stdout.isatty() or (os.name == "nt" and not os.environ.get("ANSICON")):
    try:
        # Try enabling ANSI on Windows 10+
        import ctypes
        windll = cast(Any, getattr(ctypes, "windll", None))
        if windll is not None:
            kernel32 = windll.kernel32
            kernel32.SetConsoleMode(kernel32.GetStdHandle(-11), 7)
        else:
            Colors.disable()
    except Exception:
        Colors.disable()


def state_color(state: str) -> str:
    """Return colored state string."""
    colors = {
        "pending": Colors.YELLOW,
        "running": Colors.BLUE,
        "completed": Colors.GREEN,
        "failed": Colors.RED,
        "timed_out": Colors.RED,
        "cancelled": Colors.DIM,
        "cancelling": Colors.YELLOW,
    }
    color = colors.get(state, "")
    return f"{color}{state}{Colors.RESET}"


def format_duration(td: timedelta | None) -> str:
    """Format a timedelta as human-readable string."""
    if td is None:
        return "-"
    total_seconds = int(td.total_seconds())
    if total_seconds < 60:
        return f"{total_seconds}s"
    elif total_seconds < 3600:
        mins, secs = divmod(total_seconds, 60)
        return f"{mins}m {secs}s"
    else:
        hours, remainder = divmod(total_seconds, 3600)
        mins, secs = divmod(remainder, 60)
        return f"{hours}h {mins}m"


def format_time(dt: datetime | None) -> str:
    """Format datetime for display."""
    if dt is None:
        return "-"
    return dt.strftime("%Y-%m-%d %H:%M:%S")


def format_time_short(dt: datetime | None) -> str:
    """Format datetime with just time for recent entries."""
    if dt is None:
        return "-"
    return dt.strftime("%H:%M:%S")


def truncate(s: str, length: int) -> str:
    """Truncate string with ellipsis."""
    if len(s) <= length:
        return s
    return s[:length - 3] + "..."


# ============================================================================
# EXECUTION COMMANDS
# ============================================================================

async def list_executions(executor: Stent, args):
    """List executions with filtering."""
    state_filter = args.state.lower() if args.state else None
    executions = await executor.list_executions(limit=args.limit, state=state_filter)
    
    if not executions:
        print(f"{Colors.DIM}No executions found.{Colors.RESET}")
        return

    # Header
    print(f"{Colors.BOLD}{'ID':<36} {'State':<12} {'Queue':<10} {'Started':<20} {'Duration':<10}{Colors.RESET}")
    print("-" * 90)
    
    for exc in executions:
        duration = None
        if exc.started_at:
            end_time = exc.completed_at or datetime.now()
            duration = end_time - exc.started_at
        
        queue = exc.queue or "default"
        started = format_time(exc.started_at)
        dur_str = format_duration(duration)
        
        print(f"{exc.id:<36} {state_color(exc.state):<22} {queue:<10} {started:<20} {dur_str:<10}")


async def show_execution(executor: Stent, args):
    """Show detailed execution info."""
    try:
        state = await executor.state_of(args.id)
    except ValueError:
        print(f"{Colors.RED}Execution {args.id} not found.{Colors.RESET}")
        return 1

    # Header
    print(f"\n{Colors.BOLD}Execution: {state.id}{Colors.RESET}")
    print("=" * 60)
    
    # Basic info
    print(f"  State:      {state_color(state.state)}")
    print(f"  Queue:      {state.queue or 'default'}")
    print(f"  Priority:   {state.priority}")
    if state.tags:
        print(f"  Tags:       {', '.join(state.tags)}")

    if state.counters:
        print(f"\n{Colors.BOLD}Counters{Colors.RESET}")
        for key in sorted(state.counters.keys()):
            print(f"  {key}: {state.counters[key]}")

    if state.custom_state:
        print(f"\n{Colors.BOLD}Custom State{Colors.RESET}")
        for key in sorted(state.custom_state.keys()):
            value_str = str(state.custom_state[key])
            if len(value_str) > 200:
                value_str = value_str[:200] + "..."
            print(f"  {key}: {value_str}")
    
    # Timing
    print(f"\n{Colors.BOLD}Timing{Colors.RESET}")
    print(f"  Started:    {format_time(state.started_at)}")
    print(f"  Completed:  {format_time(state.completed_at)}")
    if state.started_at:
        end_time = state.completed_at or datetime.now()
        duration = end_time - state.started_at
        print(f"  Duration:   {format_duration(duration)}")
    
    # Progress
    if state.progress:
        print(f"\n{Colors.BOLD}Progress ({len(state.progress)} steps){Colors.RESET}")
        for p in state.progress:
            timestamp = p.completed_at or p.started_at
            ts_str = format_time_short(timestamp)
            
            if p.status == "completed":
                icon = f"{Colors.GREEN}✓{Colors.RESET}"
            elif p.status == "failed":
                icon = f"{Colors.RED}✗{Colors.RESET}"
            elif p.status == "running":
                icon = f"{Colors.BLUE}▶{Colors.RESET}"
            elif p.status == "cache_hit":
                icon = f"{Colors.CYAN}⚡{Colors.RESET}"
            else:
                icon = f"{Colors.YELLOW}○{Colors.RESET}"
            
            step_name = truncate(p.step, 40)
            print(f"  [{ts_str}] {icon} {step_name}")
            if p.detail:
                print(f"             {Colors.DIM}{p.detail}{Colors.RESET}")
    
    # Result
    if state.result is not None:
        print(f"\n{Colors.BOLD}Result{Colors.RESET}")
        if state.result.ok:
            result_str = str(state.result.value)
            if len(result_str) > 200:
                result_str = result_str[:200] + "..."
            print(f"  {Colors.GREEN}OK:{Colors.RESET} {result_str}")
        else:
            error_str = str(state.result.error)
            if len(error_str) > 200:
                error_str = error_str[:200] + "..."
            print(f"  {Colors.RED}Error:{Colors.RESET} {error_str}")
    
    print()
    return 0


# ============================================================================
# TASK COMMANDS
# ============================================================================

async def list_tasks(executor: Stent, args):
    """List tasks with filtering."""
    tasks = await executor.backend.list_tasks(
        limit=args.limit,
        state=args.state.lower() if args.state else None
    )
    
    if not tasks:
        print(f"{Colors.DIM}No tasks found.{Colors.RESET}")
        return

    print(f"{Colors.BOLD}{'Task ID':<36} {'State':<10} {'Kind':<12} {'Step':<25} {'Queue':<10}{Colors.RESET}")
    print("-" * 100)
    
    for task in tasks:
        step = truncate(task.step_name, 25)
        queue = task.queue or "default"
        print(f"{task.id:<36} {state_color(task.state):<20} {task.kind:<12} {step:<25} {queue:<10}")


async def show_task(executor: Stent, args):
    """Show detailed task info."""
    task = await executor.backend.get_task(args.id)
    if not task:
        print(f"{Colors.RED}Task {args.id} not found.{Colors.RESET}")
        return 1

    print(f"\n{Colors.BOLD}Task: {task.id}{Colors.RESET}")
    print("=" * 60)
    
    print(f"  Execution:  {task.execution_id}")
    print(f"  Step:       {task.step_name}")
    print(f"  Kind:       {task.kind}")
    print(f"  State:      {state_color(task.state)}")
    print(f"  Queue:      {task.queue or 'default'}")
    print(f"  Priority:   {task.priority}")
    print(f"  Retries:    {task.retries}")
    
    if task.tags:
        print(f"  Tags:       {', '.join(task.tags)}")
    if task.worker_id:
        print(f"  Worker:     {task.worker_id}")
    if task.idempotency_key:
        print(f"  Idemp Key:  {task.idempotency_key}")
    
    print(f"\n{Colors.BOLD}Timing{Colors.RESET}")
    print(f"  Created:    {format_time(task.created_at)}")
    print(f"  Started:    {format_time(task.started_at)}")
    print(f"  Completed:  {format_time(task.completed_at)}")
    if task.scheduled_for:
        print(f"  Scheduled:  {format_time(task.scheduled_for)}")
    if task.lease_expires_at:
        print(f"  Lease exp:  {format_time(task.lease_expires_at)}")
    
    if task.retry_policy:
        rp = task.retry_policy
        print(f"\n{Colors.BOLD}Retry Policy{Colors.RESET}")
        print(f"  Max attempts: {rp.max_attempts}")
        print(f"  Backoff:      {rp.initial_delay}s × {rp.backoff_factor} (max {rp.max_delay}s)")
    
    print()
    return 0


# ============================================================================
# STATS COMMAND
# ============================================================================

async def show_stats(executor: Stent, args):
    """Show queue and execution statistics."""
    # Count executions by state
    exec_states = ["pending", "running", "completed", "failed", "timed_out", "cancelled"]
    exec_counts: dict[str, int] = {}
    for state in exec_states:
        exec_counts[state] = await executor.backend.count_executions(state=state)
    
    # Count tasks by state
    task_states = ["pending", "running", "completed", "failed"]
    task_counts: dict[str, int] = {}
    for state in task_states:
        task_counts[state] = await executor.backend.count_tasks(state=state)
    
    # Queue depths
    pending_tasks = await executor.backend.list_tasks(limit=1000, state="pending")
    queues: dict[str, int] = {}
    for t in pending_tasks:
        q = t.queue or "default"
        queues[q] = queues.get(q, 0) + 1
    
    # DLQ count
    dlq_count = await executor.backend.count_dead_tasks()
    
    # Running activities
    running_tasks = await executor.get_running_activities()
    
    # Print stats
    print(f"\n{Colors.BOLD}Stent Statistics{Colors.RESET}")
    print("=" * 50)
    
    print(f"\n{Colors.BOLD}Executions{Colors.RESET}")
    total_exec = sum(exec_counts.values())
    print(f"  Total:     {total_exec}")
    for state, count in exec_counts.items():
        if count > 0:
            print(f"  {state.capitalize():<10} {count}")
    
    print(f"\n{Colors.BOLD}Tasks{Colors.RESET}")
    total_tasks = sum(task_counts.values())
    print(f"  Total:     {total_tasks}")
    for state, count in task_counts.items():
        if count > 0:
            print(f"  {state.capitalize():<10} {count}")
    
    print(f"\n{Colors.BOLD}Queue Depths{Colors.RESET}")
    if queues:
        for queue, depth in sorted(queues.items()):
            bar = "█" * min(depth, 30)
            print(f"  {queue:<15} {depth:>5} {Colors.CYAN}{bar}{Colors.RESET}")
    else:
        print(f"  {Colors.DIM}All queues empty{Colors.RESET}")
    
    print(f"\n{Colors.BOLD}Dead Letter Queue{Colors.RESET}")
    print(f"  Count:     {dlq_count}")
    
    if running_tasks:
        print(f"\n{Colors.BOLD}Running Tasks ({len(running_tasks)}){Colors.RESET}")
        for t in running_tasks[:10]:
            step = truncate(t.step_name, 30)
            worker = t.worker_id or "unknown"
            print(f"  {step:<30} worker={worker}")
        if len(running_tasks) > 10:
            print(f"  ... and {len(running_tasks) - 10} more")
    
    print()


# ============================================================================
# DLQ COMMANDS
# ============================================================================

async def dlq_list(executor: Stent, args):
    """List dead-letter queue entries."""
    records = await executor.list_dead_letters(limit=args.limit)
    if not records:
        print(f"{Colors.DIM}Dead-letter queue is empty.{Colors.RESET}")
        return

    print(f"{Colors.BOLD}{'Task ID':<36} {'Step':<25} {'Moved At':<20} {'Reason':<30}{Colors.RESET}")
    print("-" * 115)
    
    for record in records:
        task = record.task
        step = truncate(task.step_name, 25)
        moved = format_time(record.moved_at)
        reason = truncate(record.reason or "", 30)
        print(f"{task.id:<36} {step:<25} {moved:<20} {reason:<30}")


async def dlq_show(executor: Stent, args):
    """Show dead-letter entry details."""
    record = await executor.get_dead_letter(args.id)
    if not record:
        print(f"{Colors.RED}Dead-letter task {args.id} not found.{Colors.RESET}")
        return 1

    task = record.task
    print(f"\n{Colors.BOLD}Dead-Letter Task: {task.id}{Colors.RESET}")
    print("=" * 60)
    
    print(f"  Execution:  {task.execution_id}")
    print(f"  Step:       {task.step_name}")
    print(f"  Kind:       {task.kind}")
    print(f"  Queue:      {task.queue or 'default'}")
    print(f"  Retries:    {task.retries}")
    
    if task.tags:
        print(f"  Tags:       {', '.join(task.tags)}")
    
    print(f"\n{Colors.BOLD}Dead-Letter Info{Colors.RESET}")
    print(f"  Moved At:   {format_time(record.moved_at)}")
    print(f"  Reason:     {record.reason}")
    
    print()
    return 0


async def dlq_replay(executor: Stent, args):
    """Replay a dead-letter entry."""
    try:
        new_id = await executor.replay_dead_letter(args.id, queue=args.queue)
        print(f"{Colors.GREEN}✓{Colors.RESET} Replayed dead-letter task {args.id}")
        print(f"  New task ID: {new_id}")
    except ValueError as e:
        print(f"{Colors.RED}Error:{Colors.RESET} {e}")
        return 1


async def dlq_discard(executor: Stent, args):
    """Discard a dead-letter entry."""
    if not args.force:
        confirm = input(f"Discard dead-letter task {args.id}? [y/N] ")
        if confirm.lower() != "y":
            print("Cancelled.")
            return

    success = await executor.discard_dead_letter(args.id)
    if success:
        print(f"{Colors.GREEN}✓{Colors.RESET} Discarded dead-letter task {args.id}")
    else:
        print(f"{Colors.RED}Error:{Colors.RESET} Task {args.id} not found")
        return 1


async def dlq_replay_all(executor: Stent, args):
    """Replay all dead-letter entries."""
    records = await executor.list_dead_letters(limit=1000)
    if not records:
        print(f"{Colors.DIM}Dead-letter queue is empty.{Colors.RESET}")
        return

    if not args.force:
        confirm = input(f"Replay {len(records)} dead-letter tasks? [y/N] ")
        if confirm.lower() != "y":
            print("Cancelled.")
            return

    replayed = 0
    for record in records:
        try:
            await executor.replay_dead_letter(record.task.id)
            replayed += 1
        except Exception as e:
            print(f"{Colors.RED}Error replaying {record.task.id}:{Colors.RESET} {e}")

    print(f"{Colors.GREEN}✓{Colors.RESET} Replayed {replayed}/{len(records)} tasks")


# ============================================================================
# CLEANUP COMMAND
# ============================================================================

async def cleanup(executor: Stent, args):
    """Clean up old executions and dead-letter entries."""
    cutoff = datetime.now() - timedelta(days=args.days)
    
    if args.dry_run:
        print(f"Dry run: Would clean up records older than {format_time(cutoff)}")
        return

    if not args.force:
        confirm = input(f"Delete records older than {args.days} days? [y/N] ")
        if confirm.lower() != "y":
            print("Cancelled.")
            return

    exec_count = 0
    dlq_count = 0
    
    if args.all or args.executions:
        exec_count = await executor.backend.cleanup_executions(cutoff)
        print(f"  Cleaned up {exec_count} executions")
    
    if args.all or args.dlq:
        dlq_count = await executor.backend.cleanup_dead_letters(cutoff)
        print(f"  Cleaned up {dlq_count} dead-letter entries")
    
    total = exec_count + dlq_count
    print(f"\n{Colors.GREEN}✓{Colors.RESET} Cleaned up {total} total records")


# ============================================================================
# SIGNAL COMMAND
# ============================================================================

async def send_signal(executor: Stent, args):
    """Send a signal to an execution."""
    # Parse payload
    payload = args.payload
    if args.json:
        try:
            payload = json.loads(payload)
        except json.JSONDecodeError as e:
            print(f"{Colors.RED}Invalid JSON:{Colors.RESET} {e}")
            return 1

    try:
        await executor.send_signal(args.execution_id, args.name, payload)
        print(f"{Colors.GREEN}✓{Colors.RESET} Sent signal '{args.name}' to execution {args.execution_id}")
    except Exception as e:
        print(f"{Colors.RED}Error:{Colors.RESET} {e}")
        return 1


# ============================================================================
# WATCH COMMAND
# ============================================================================

async def _watch_simple(executor: Stent, args):
    """Simple fallback watch mode without rich."""
    print(f"{Colors.BOLD}Stent Watch Mode{Colors.RESET} (press Ctrl+C to exit)")
    print("-" * 50)
    print(f"{Colors.DIM}Tip: Install 'rich' for a better experience{Colors.RESET}\n")
    
    try:
        while True:
            # Clear screen (basic)
            if os.name == "nt":
                os.system("cls")
            else:
                print("\033[2J\033[H", end="")
            
            print(f"{Colors.BOLD}Stent Watch{Colors.RESET} - {datetime.now().strftime('%H:%M:%S')}")
            print("=" * 50)
            
            # Quick stats
            pending = await executor.backend.count_tasks(state="pending")
            running = await executor.backend.count_tasks(state="running")
            dlq = await executor.backend.count_dead_tasks()
            
            print(f"\nPending: {Colors.YELLOW}{pending}{Colors.RESET}  "
                  f"Running: {Colors.BLUE}{running}{Colors.RESET}  "
                  f"DLQ: {Colors.RED}{dlq}{Colors.RESET}")
            
            # Recent executions
            print(f"\n{Colors.BOLD}Recent Executions{Colors.RESET}")
            executions = await executor.list_executions(limit=10)
            for exc in executions:
                print(f"  {exc.id[:8]}... {state_color(exc.state)}")
            
            print(f"\n{Colors.DIM}Refreshing in {args.interval}s...{Colors.RESET}")
            await asyncio.sleep(args.interval)
            
    except KeyboardInterrupt:
        print("\nExiting watch mode.")


async def _watch_rich(executor: Stent, args):
    """Rich-based watch mode with fancy UI."""
    from rich.console import Console  # type: ignore
    from rich.live import Live  # type: ignore
    from rich.table import Table  # type: ignore
    from rich.panel import Panel  # type: ignore
    from rich.layout import Layout  # type: ignore
    
    console = Console()
    
    def make_layout():
        layout = Layout()
        layout.split_column(
            Layout(name="header", size=3),
            Layout(name="main"),
            Layout(name="footer", size=3),
        )
        layout["main"].split_row(
            Layout(name="stats", ratio=1),
            Layout(name="executions", ratio=2),
        )
        return layout
    
    async def generate_display():
        # Stats
        pending = await executor.backend.count_tasks(state="pending")
        running = await executor.backend.count_tasks(state="running")
        completed = await executor.backend.count_tasks(state="completed")
        failed = await executor.backend.count_tasks(state="failed")
        dlq = await executor.backend.count_dead_tasks()
        
        stats_table = Table(show_header=False, box=None)
        stats_table.add_row("Pending", f"[yellow]{pending}[/]")
        stats_table.add_row("Running", f"[blue]{running}[/]")
        stats_table.add_row("Completed", f"[green]{completed}[/]")
        stats_table.add_row("Failed", f"[red]{failed}[/]")
        stats_table.add_row("DLQ", f"[red]{dlq}[/]")
        
        # Executions
        exec_table = Table(title="Recent Executions")
        exec_table.add_column("ID", style="dim")
        exec_table.add_column("State")
        exec_table.add_column("Started")
        
        executions = await executor.list_executions(limit=15)
        for exc in executions:
            state_style = {
                "pending": "yellow",
                "running": "blue",
                "completed": "green",
                "failed": "red",
            }.get(exc.state, "white")
            
            started = format_time_short(exc.started_at)
            exec_table.add_row(
                exc.id[:12] + "...",
                f"[{state_style}]{exc.state}[/]",
                started
            )
        
        layout = make_layout()
        layout["header"].update(Panel(f"[bold]Stent Watch[/] - {datetime.now().strftime('%H:%M:%S')}"))
        layout["stats"].update(Panel(stats_table, title="Stats"))
        layout["executions"].update(exec_table)
        layout["footer"].update(Panel("[dim]Press Ctrl+C to exit[/]"))
        
        return layout
    
    try:
        with Live(await generate_display(), console=console, refresh_per_second=1) as live:
            while True:
                await asyncio.sleep(args.interval)
                live.update(await generate_display())
    except KeyboardInterrupt:
        console.print("\n[dim]Exiting watch mode.[/]")


async def watch(executor: Stent, args):
    """Live monitoring of executions and queues."""
    try:
        import rich  # type: ignore  # noqa: F401
        has_rich = True
    except ImportError:
        has_rich = False
    
    if has_rich:
        await _watch_rich(executor, args)
    else:
        await _watch_simple(executor, args)


# ============================================================================
# MAIN
# ============================================================================

async def main_async():
    parser = argparse.ArgumentParser(
        description="Stent CLI - Manage distributed durable functions",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog="""
Examples:
  stent list                          List recent executions
  stent list --state running          List running executions
  stent show <execution-id>           Show execution details
  stent stats                         Show queue statistics
  stent dlq list                      List dead-letter queue
  stent dlq replay <task-id>          Replay failed task
  stent cleanup --days 7              Clean up old records
  stent signal <exec-id> my_signal    Send signal
  stent watch                         Live monitoring
        """
    )
    
    default_db = os.environ.get("STENT_DB", "stent.sqlite")
    parser.add_argument(
        "--db", 
        default=default_db, 
        help=f"Database path or connection string (default: {default_db}, env: STENT_DB)"
    )
    parser.add_argument(
        "--no-color",
        action="store_true",
        help="Disable colored output"
    )
    
    subparsers = parser.add_subparsers(dest="command", required=True)
    
    # list command
    list_parser = subparsers.add_parser("list", help="List executions")
    list_parser.add_argument("--limit", "-n", type=int, default=20, help="Number of executions (default: 20)")
    list_parser.add_argument("--state", "-s", help="Filter by state (pending, running, completed, failed)")
    
    # show command
    show_parser = subparsers.add_parser("show", help="Show execution details")
    show_parser.add_argument("id", help="Execution ID")
    
    # tasks command
    tasks_parser = subparsers.add_parser("tasks", help="Task operations")
    tasks_sub = tasks_parser.add_subparsers(dest="tasks_command", required=True)
    
    tasks_list = tasks_sub.add_parser("list", help="List tasks")
    tasks_list.add_argument("--limit", "-n", type=int, default=20, help="Number of tasks")
    tasks_list.add_argument("--state", "-s", help="Filter by state")
    
    tasks_show = tasks_sub.add_parser("show", help="Show task details")
    tasks_show.add_argument("id", help="Task ID")
    
    # stats command
    stats_parser = subparsers.add_parser("stats", help="Show queue statistics")
    
    # dlq command
    dlq_parser = subparsers.add_parser("dlq", help="Dead-letter queue operations")
    dlq_sub = dlq_parser.add_subparsers(dest="dlq_command", required=True)
    
    dlq_list_parser = dlq_sub.add_parser("list", help="List DLQ entries")
    dlq_list_parser.add_argument("--limit", "-n", type=int, default=20, help="Number of entries")
    
    dlq_show_parser = dlq_sub.add_parser("show", help="Show DLQ entry details")
    dlq_show_parser.add_argument("id", help="Dead-letter task ID")
    
    dlq_replay_parser = dlq_sub.add_parser("replay", help="Replay a DLQ entry")
    dlq_replay_parser.add_argument("id", help="Dead-letter task ID")
    dlq_replay_parser.add_argument("--queue", help="Override queue when replaying")
    
    dlq_discard_parser = dlq_sub.add_parser("discard", help="Discard a DLQ entry")
    dlq_discard_parser.add_argument("id", help="Dead-letter task ID")
    dlq_discard_parser.add_argument("--force", "-f", action="store_true", help="Skip confirmation")
    
    dlq_replay_all_parser = dlq_sub.add_parser("replay-all", help="Replay all DLQ entries")
    dlq_replay_all_parser.add_argument("--force", "-f", action="store_true", help="Skip confirmation")
    
    # cleanup command
    cleanup_parser = subparsers.add_parser("cleanup", help="Clean up old records")
    cleanup_parser.add_argument("--days", "-d", type=int, default=30, help="Delete records older than N days (default: 30)")
    cleanup_parser.add_argument("--executions", action="store_true", help="Clean up executions only")
    cleanup_parser.add_argument("--dlq", action="store_true", help="Clean up DLQ only")
    cleanup_parser.add_argument("--all", "-a", action="store_true", help="Clean up all (default if no specific flag)")
    cleanup_parser.add_argument("--dry-run", action="store_true", help="Show what would be deleted")
    cleanup_parser.add_argument("--force", "-f", action="store_true", help="Skip confirmation")
    
    # signal command
    signal_parser = subparsers.add_parser("signal", help="Send signal to execution")
    signal_parser.add_argument("execution_id", help="Execution ID")
    signal_parser.add_argument("name", help="Signal name")
    signal_parser.add_argument("payload", nargs="?", default="", help="Signal payload")
    signal_parser.add_argument("--json", "-j", action="store_true", help="Parse payload as JSON")
    
    # watch command
    watch_parser = subparsers.add_parser("watch", help="Live monitoring")
    watch_parser.add_argument("--interval", "-i", type=float, default=2.0, help="Refresh interval in seconds (default: 2)")
    
    args = parser.parse_args()
    
    # Handle no-color flag
    if args.no_color:
        Colors.disable()
    
    # Determine backend
    if "://" in args.db or "postgres" in args.db.lower():
        backend = Stent.backends.PostgresBackend(args.db)
    else:
        backend = Stent.backends.SQLiteBackend(args.db)

    await backend.init_db()
    executor = Stent(backend=backend)
    
    # Handle --all default for cleanup
    if args.command == "cleanup" and not args.executions and not args.dlq:
        args.all = True
    
    try:
        # Dispatch to command handlers
        result = 0
        if args.command == "list":
            await list_executions(executor, args)
        elif args.command == "show":
            result = await show_execution(executor, args) or 0
        elif args.command == "tasks":
            if args.tasks_command == "list":
                await list_tasks(executor, args)
            elif args.tasks_command == "show":
                result = await show_task(executor, args) or 0
        elif args.command == "stats":
            await show_stats(executor, args)
        elif args.command == "dlq":
            if args.dlq_command == "list":
                await dlq_list(executor, args)
            elif args.dlq_command == "show":
                result = await dlq_show(executor, args) or 0
            elif args.dlq_command == "replay":
                result = await dlq_replay(executor, args) or 0
            elif args.dlq_command == "discard":
                result = await dlq_discard(executor, args) or 0
            elif args.dlq_command == "replay-all":
                result = await dlq_replay_all(executor, args) or 0
        elif args.command == "cleanup":
            await cleanup(executor, args)
        elif args.command == "signal":
            result = await send_signal(executor, args) or 0
        elif args.command == "watch":
            await watch(executor, args)

        return result
    finally:
        await backend.close()


def main():
    try:
        result = asyncio.run(main_async())
        sys.exit(result or 0)
    except KeyboardInterrupt:
        print("\nInterrupted.")
        sys.exit(130)


if __name__ == "__main__":
    main()



import unittest
import asyncio
import os
import logging
from stent import Stent, Result
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

logging.basicConfig(level=logging.INFO)

@Stent.durable()
async def leaf_activity():
    return "done"

@Stent.durable()
async def deadlocking_orchestrator():
    # Schedules a leaf task and waits for it
    return await leaf_activity()

class TestDeadlock(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"deadlock_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)

    async def asyncTearDown(self):
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_single_concurrency_deadlock(self):
        # We start a worker with concurrency=1
        # If the orchestrator holds the slot while waiting, the leaf activity can never run.
        worker_task = asyncio.create_task(self.executor.serve(max_concurrency=1, poll_interval=0.1))
        
        try:
            exec_id = await self.executor.dispatch(deadlocking_orchestrator)
            
            # If fix works, this returns. If not, it times out.
            # We set a generous expiry for the test
            result = await self.executor.wait_for(exec_id, expiry=30.0)
            
            self.assertTrue(result.ok)
            self.assertEqual(result.value, "done")
            
        finally:
            worker_task.cancel()
            try:
                await worker_task
            except asyncio.CancelledError:
                pass
            await self.executor.shutdown()




import unittest
import asyncio
import os
import logging
import contextlib
from stent import Stent, Result, sleep
from stent.executor import _original_sleep
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

logger = logging.getLogger(__name__)

@Stent.durable()
async def sleep_task():
    await sleep("0.2s")
    return "done"

@Stent.durable()
async def noop_task():
    pass

class TestHelpers(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"helpers_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend) # Ensure clean slate
        self.executor = Stent(backend=self.backend)
        self.worker_task = asyncio.create_task(self.executor.serve(poll_interval=0.1))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_asyncio_sleep_patched_but_transparent_outside_context(self):
        """asyncio.sleep is patched but behaves normally outside durable context."""
        self.assertIsNot(asyncio.sleep, _original_sleep)

        observed = []

        async def third_party_coroutine():
            await asyncio.sleep(0.01)
            observed.append("ok")

        await third_party_coroutine()
        self.assertEqual(observed, ["ok"])

    async def test_queue_depth(self):
        # Cancel default worker to prevent it from consuming tasks
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
            
        self.assertEqual(await self.executor.queue_depth(), 0)
        
        # Dispatch a task
        await self.executor.dispatch(noop_task)
        self.assertEqual(await self.executor.queue_depth(), 1) 
        
        # Dispatch another
        await self.executor.dispatch(noop_task, queue="special")
        self.assertEqual(await self.executor.queue_depth(), 2)
        self.assertEqual(await self.executor.queue_depth(queue="special"), 1)

    async def test_list_executions(self):
        ids = []
        for _ in range(5):
            ids.append(await self.executor.dispatch(noop_task))
            
        executions = await self.executor.list_executions(limit=10)
        self.assertEqual(len(executions), 5)
        self.assertEqual(executions[0].state, "pending")
        
        # Start worker to finish them
        task = asyncio.create_task(self.executor.serve(poll_interval=0.1))
        
        for eid in ids:
            await self.executor.wait_for(eid, expiry=2.0)
            
        executions = await self.executor.list_executions(state="completed")
        self.assertEqual(len(executions), 5)
        
        task.cancel()
        try:
            await task
        except asyncio.CancelledError:
            pass

    async def test_worker_lifecycle_drain_and_status(self):
        # Stop default worker to avoid interference
        self.worker_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await self.worker_task

        lifecycle = self.executor.create_worker_lifecycle(name="probe")
        worker = asyncio.create_task(self.executor.serve(poll_interval=0.05, lifecycle=lifecycle))
        await lifecycle.wait_until_ready()
        status = self.executor.worker_status_overview()
        self.assertTrue(status["ready"])
        self.assertFalse(status["draining"])

        self.executor.request_worker_drain(lifecycle)
        await lifecycle.wait_until_stopped()
        await worker
        stopped_status = self.executor.worker_status_overview()
        self.assertTrue(stopped_status["draining"] or not stopped_status["workers"])



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast
import functools
import logging
from stent import Stent

logger = logging.getLogger(__name__)

_HAS_OTEL = False

# Try to import opentelemetry at runtime
try:
    from opentelemetry import trace as _trace  # type: ignore[import-not-found]
    from opentelemetry.trace import Status as _Status, StatusCode as _StatusCode  # type: ignore[import-not-found]
    _HAS_OTEL = True
except ImportError:
    # Provide stub values for when otel is not installed
    _trace = None
    _Status = None
    _StatusCode = None

# Type hints for when type checking (opentelemetry may not be installed at runtime)
if TYPE_CHECKING:
    from opentelemetry import trace  # type: ignore[import-not-found]
    from opentelemetry.trace import Status, StatusCode, Tracer  # type: ignore[import-not-found]
    # Create type aliases that are only used during type checking
    TraceModule = Any
    StatusType = Any
    StatusCodeType = Any

F = TypeVar('F', bound=Callable[..., Any])


def instrument(tracer_provider: Any = None) -> bool:
    """
    Instruments the Stent library with OpenTelemetry.
    Returns True if instrumentation was installed, False otherwise.
    """
    if not _HAS_OTEL or _trace is None:
        logger.warning("OpenTelemetry not installed; skipping Stent instrumentation.")
        return False

    tracer = _trace.get_tracer("stent", tracer_provider=tracer_provider)
    
    _instrument_executor(tracer)
    return True


def _instrument_executor(tracer: Any) -> None:
    """
    Internal function to instrument the executor methods.
    
    Note: This function is only called when _HAS_OTEL is True,
    so _trace, _Status, and _StatusCode are guaranteed to be non-None.
    """
    # Idempotency check
    if getattr(Stent.dispatch, "_is_otel_instrumented", False):
        return

    # Local references to ensure type checker knows these are non-None
    # (they are guaranteed to be set because instrument() guards this call)
    trace_mod = cast(Any, _trace)
    status_cls = cast(Any, _Status)
    status_code_cls = cast(Any, _StatusCode)

    original_dispatch = Stent.dispatch
    original_handle_task = Stent._handle_task
    
    @functools.wraps(original_dispatch)
    async def dispatch_wrapper(self: Stent, fn: Any, *args: Any, **kwargs: Any) -> str:
        # Resolve name properly if it's a wrapped function
        name = "unknown"
        if hasattr(fn, "__name__"):
            name = fn.__name__
        
        with tracer.start_as_current_span(f"stent.dispatch {name}", kind=trace_mod.SpanKind.PRODUCER) as span:
            span.set_attribute("stent.function", name)
            
            try:
                exec_id = await original_dispatch(self, fn, *args, **kwargs)
                span.set_attribute("stent.execution_id", exec_id)
                return exec_id
            except Exception as e:
                span.record_exception(e)
                span.set_status(status_cls(status_code_cls.ERROR, str(e)))
                raise

    setattr(dispatch_wrapper, "_is_otel_instrumented", True)
    Stent.dispatch = dispatch_wrapper  # type: ignore[method-assign]
    
    @functools.wraps(original_handle_task)
    async def handle_task_wrapper(self: Stent, task: Any, worker_id: str, *args: Any, **kwargs: Any) -> None:
        # Consumer span
        with tracer.start_as_current_span(f"stent.execute {task.step_name}", kind=trace_mod.SpanKind.CONSUMER) as span:
             span.set_attribute("stent.task_id", task.id)
             span.set_attribute("stent.execution_id", task.execution_id)
             span.set_attribute("stent.step", task.step_name)
             span.set_attribute("stent.worker_id", worker_id)
             
             try:
                 await original_handle_task(self, task, worker_id, *args, **kwargs)
                 
                 # Check task status after execution
                 if task.state == "failed":
                     span.set_status(status_cls(status_code_cls.ERROR, str(task.error)))
                 elif task.state == "completed":
                     span.set_status(status_cls(status_code_cls.OK))
                     
             except Exception as e:
                 span.record_exception(e)
                 span.set_status(status_cls(status_code_cls.ERROR, str(e)))
                 raise

    setattr(handle_task_wrapper, "_is_otel_instrumented", True)
    Stent._handle_task = handle_task_wrapper  # type: ignore[method-assign]



import asyncpg
import asyncpg.pool
import json
import logging
from datetime import datetime, timedelta
from typing import List, Optional, Any, Union
from stent.backend.base import Backend
from stent.core import ExecutionRecord, TaskRecord, ExecutionProgress, RetryPolicy, SignalRecord, DeadLetterRecord
from stent.backend.utils import (
    build_filtered_count_query,
    build_filtered_list_query,
    dollar_placeholder,
    execution_row_values,
    row_to_dead_letter,
    row_to_execution,
    row_to_progress,
    row_to_signal,
    row_to_task,
    retry_policy_from_json,
    retry_policy_to_json,
    task_record_to_json,
    task_row_values,
)

logger = logging.getLogger(__name__)

# Type alias for connection objects (both direct connections and pool proxies)
# Using Any because PoolConnectionProxy and Connection both have execute() but type checker doesn't know
_Conn = Any

class PostgresBackend(Backend):
    def __init__(self, dsn: str, min_pool_size: int = 2, max_pool_size: int = 10):
        """
        Initialize Postgres backend.
        
        Args:
            dsn: PostgreSQL connection string
            min_pool_size: Minimum connections in pool (default 2)
            max_pool_size: Maximum connections in pool (default 10)
        """
        self.dsn = dsn
        self._min_pool_size = min_pool_size
        self._max_pool_size = max_pool_size
        self.pool: Optional[asyncpg.Pool] = None
        self._closed = False

    async def close(self) -> None:
        """Close the connection pool and release resources."""
        self._closed = True
        if self.pool is not None:
            await self.pool.close()
            self.pool = None
            logger.info("Postgres connection pool closed")

    async def init_db(self):
        if self._closed:
            raise RuntimeError("PostgresBackend has been closed")
        if not self.pool:
            self.pool = await asyncpg.create_pool(
                self.dsn,
                min_size=self._min_pool_size,
                max_size=self._max_pool_size,
            )
        
        assert self.pool is not None
        
        async with self.pool.acquire() as conn:
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS signals (
                    execution_id TEXT,
                    name TEXT,
                    payload BYTEA,
                    created_at TIMESTAMP,
                    consumed BOOLEAN,
                    consumed_at TIMESTAMP,
                    PRIMARY KEY (execution_id, name)
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS executions (
                    id TEXT PRIMARY KEY,
                    root_function TEXT,
                    state TEXT,
                    args BYTEA,
                    kwargs BYTEA,
                    result BYTEA,
                    error BYTEA,
                    retries INTEGER,
                    created_at TIMESTAMP,
                    started_at TIMESTAMP,
                    completed_at TIMESTAMP,
                    expiry_at TIMESTAMP,
                    tags TEXT,
                    priority INTEGER,
                    queue TEXT
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS execution_progress (
                    execution_id TEXT,
                    step TEXT,
                    status TEXT,
                    started_at TIMESTAMP,
                    completed_at TIMESTAMP,
                    detail TEXT,
                    ordinal SERIAL PRIMARY KEY
                )
            """)
            await conn.execute("CREATE INDEX IF NOT EXISTS idx_progress_exec ON execution_progress(execution_id)")
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS tasks (
                    id TEXT PRIMARY KEY,
                    execution_id TEXT,
                    step_name TEXT,
                    kind TEXT,
                    parent_task_id TEXT,
                    state TEXT,
                    args BYTEA,
                    kwargs BYTEA,
                    result BYTEA,
                    error BYTEA,
                    retries INTEGER,
                    created_at TIMESTAMP,
                    started_at TIMESTAMP,
                    completed_at TIMESTAMP,
                    worker_id TEXT,
                    lease_expires_at TIMESTAMP,
                    tags TEXT,
                    priority INTEGER,
                    queue TEXT,
                    idempotency_key TEXT,
                    retry_policy TEXT,
                    scheduled_for TIMESTAMP
                )
            """)
            await conn.execute("CREATE INDEX IF NOT EXISTS idx_tasks_state_queue_scheduled ON tasks(state, queue, scheduled_for)")
            await conn.execute("CREATE INDEX IF NOT EXISTS idx_tasks_priority_created ON tasks(priority, created_at)")
            await conn.execute("CREATE INDEX IF NOT EXISTS idx_tasks_execution ON tasks(execution_id)")
            await conn.execute("CREATE INDEX IF NOT EXISTS idx_tasks_step_lease ON tasks(step_name, state, lease_expires_at)")
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS dead_tasks (
                    id TEXT PRIMARY KEY,
                    reason TEXT,
                    moved_at TIMESTAMP,
                    data TEXT -- full JSON dump of task
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS cache (
                    key TEXT PRIMARY KEY,
                    value BYTEA,
                    expires_at TIMESTAMP
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS idempotency (
                    key TEXT PRIMARY KEY,
                    value BYTEA
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS execution_counters (
                    execution_id TEXT,
                    name TEXT,
                    value DOUBLE PRECISION NOT NULL,
                    PRIMARY KEY (execution_id, name)
                )
            """)
            await conn.execute("""
                CREATE TABLE IF NOT EXISTS execution_state (
                    execution_id TEXT,
                    key TEXT,
                    value BYTEA,
                    PRIMARY KEY (execution_id, key)
                )
            """)

    def _execution_row_values(self, record: ExecutionRecord) -> tuple[Any, ...]:
        return execution_row_values(record)

    async def _insert_execution(self, conn: _Conn, record: ExecutionRecord) -> None:
        await conn.execute(
            "INSERT INTO executions (id, root_function, state, args, kwargs, result, error, retries, created_at, started_at, completed_at, expiry_at, tags, priority, queue) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15)",
            *self._execution_row_values(record),
        )
        for p in record.progress:
            await conn.execute(
                "INSERT INTO execution_progress (execution_id, step, status, started_at, completed_at, detail) VALUES ($1, $2, $3, $4, $5, $6)",
                record.id,
                p.step,
                p.status,
                p.started_at,
                p.completed_at,
                p.detail,
            )

    def _task_row_values(self, task: TaskRecord) -> tuple[Any, ...]:
        return task_row_values(task)

    async def _insert_task(self, conn: _Conn, task: TaskRecord) -> None:
        await conn.execute(
            "INSERT INTO tasks (id, execution_id, step_name, kind, parent_task_id, state, args, kwargs, result, error, retries, created_at, started_at, completed_at, worker_id, lease_expires_at, tags, priority, queue, idempotency_key, retry_policy, scheduled_for) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22)",
            *self._task_row_values(task),
        )

    async def create_execution(self, record: ExecutionRecord) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await self._insert_execution(conn, record)

    async def create_execution_with_root_task(self, record: ExecutionRecord, task: TaskRecord) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                await self._insert_execution(conn, record)
                await self._insert_task(conn, task)

    async def get_execution(self, execution_id: str) -> ExecutionRecord | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow("SELECT * FROM executions WHERE id = $1", execution_id)
            if not row:
                return None
            
            # Fetch progress
            progress = []
            p_rows = await conn.fetch("SELECT * FROM execution_progress WHERE execution_id = $1 ORDER BY ordinal", execution_id)
            for pr in p_rows:
                 progress.append(self._row_to_progress(pr))

            return self._row_to_execution(row, progress)

    async def update_execution(self, record: ExecutionRecord) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.execute("""
                UPDATE executions SET
                    state=$1, args=$2, kwargs=$3, result=$4, error=$5, retries=$6,
                    started_at=$7, completed_at=$8, expiry_at=$9, tags=$10,
                    priority=$11, queue=$12
                WHERE id=$13
            """, 
                record.state, record.args, record.kwargs, record.result, record.error,
                record.retries, record.started_at, record.completed_at, record.expiry_at,
                json.dumps(record.tags), record.priority, record.queue, record.id
            )

    async def list_executions(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[ExecutionRecord]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            query, params = build_filtered_list_query(
                table="executions",
                filters=[("state", state)],
                order_by="created_at DESC",
                limit=limit,
                offset=offset,
                placeholder=dollar_placeholder,
            )
            
            rows = await conn.fetch(query, *params)
            results = []
            for row in rows:
                results.append(self._row_to_execution(row, progress=[]))
            return results

    async def count_executions(self, state: str | None = None) -> int:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            query, params = build_filtered_count_query(
                table="executions",
                filters=[("state", state)],
                placeholder=dollar_placeholder,
            )

            val = await conn.fetchval(query, *params)
            return val if val else 0

    async def create_task(self, task: TaskRecord) -> None:
        await self.create_tasks([task])

    async def create_tasks(self, tasks: List[TaskRecord]) -> None:
        if not tasks:
            return
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.executemany(
                "INSERT INTO tasks (id, execution_id, step_name, kind, parent_task_id, state, args, kwargs, result, error, retries, created_at, started_at, completed_at, worker_id, lease_expires_at, tags, priority, queue, idempotency_key, retry_policy, scheduled_for) VALUES ($1, $2, $3, $4, $5, $6, $7, $8, $9, $10, $11, $12, $13, $14, $15, $16, $17, $18, $19, $20, $21, $22)",
                [self._task_row_values(task) for task in tasks],
            )

    async def count_tasks(self, queue: str | None = None, state: str | None = None) -> int:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            query, params = build_filtered_count_query(
                table="tasks",
                filters=[("queue", queue), ("state", state)],
                placeholder=dollar_placeholder,
            )
            
            val = await conn.fetchval(query, *params)
            return val if val else 0

    async def get_task(self, task_id: str) -> TaskRecord | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow("SELECT * FROM tasks WHERE id = $1", task_id)
            if not row:
                return None
            return self._row_to_task(row)

    async def update_task(self, task: TaskRecord) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.execute("""
                UPDATE tasks SET
                    state=$1, result=$2, error=$3, retries=$4, started_at=$5, completed_at=$6,
                    worker_id=$7, lease_expires_at=$8, scheduled_for=$9
                WHERE id=$10
            """, 
                task.state, task.result, task.error, task.retries, task.started_at,
                task.completed_at, task.worker_id, task.lease_expires_at, task.scheduled_for, task.id
            )

    async def list_tasks(self, limit: int = 10, offset: int = 0, state: str | None = None) -> List[TaskRecord]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            query, params = build_filtered_list_query(
                table="tasks",
                filters=[("state", state)],
                order_by="created_at DESC",
                limit=limit,
                offset=offset,
                placeholder=dollar_placeholder,
            )
            
            rows = await conn.fetch(query, *params)
            return [self._row_to_task(row) for row in rows]

    async def claim_next_task(
        self,
        *,
        worker_id: str,
        queues: List[str] | None = None,
        tags: List[str] | None = None,
        now: datetime | None = None,
        lease_duration: timedelta | None = None,
        concurrency_limits: dict[str, int] | None = None,
    ) -> TaskRecord | None:
        if now is None:
            now = datetime.now()
        if lease_duration is None:
            lease_duration = timedelta(minutes=5)
            
        expires_at = now + lease_duration
        worker_tags_set = set(tags or [])
        should_filter_by_tags = bool(tags)
        
        # Helper for queues condition
        queue_clause = ""
        # $1=now, $2=now
        params: List[Any] = [now, now]
        
        if queues:
            placeholders: List[str] = []
            for q in queues:
                params.append(q)
                placeholders.append(f"${len(params)}")
            
            queue_clause = f"AND (queue IN ({', '.join(placeholders)}) OR queue IS NULL)"
        else:
            queue_clause = "AND 1=1"

        assert self.pool is not None
        async with self.pool.acquire() as conn:
            async with conn.transaction():
                # 1. Fetch candidates
                query = f"""
                    SELECT * FROM tasks
                    WHERE (
                        state='pending'
                        OR (state='running' AND lease_expires_at < $1)
                    )
                    AND (scheduled_for IS NULL OR scheduled_for <= $2)
                    AND kind != 'signal'
                    {queue_clause}
                    ORDER BY priority DESC, created_at ASC
                    LIMIT 50
                    FOR UPDATE SKIP LOCKED
                """
                rows = await conn.fetch(query, *params)
                
                if not rows:
                    return None
                
                for row in rows:
                    if should_filter_by_tags:
                        task_tags = json.loads(row["tags"]) if row["tags"] else []
                        if task_tags and worker_tags_set.isdisjoint(task_tags):
                            continue

                    step_name = row["step_name"]
                    limit = concurrency_limits.get(step_name) if concurrency_limits else None
                    
                    if limit is not None:
                         # Check count of currently running tasks for this step
                         # We exclude lease-expired ones from "active" count implicitly by checking lease_expires_at > now
                         count_val = await conn.fetchval("""
                            SELECT COUNT(*) FROM tasks 
                            WHERE step_name = $1 
                            AND state = 'running' 
                            AND lease_expires_at > $2
                         """, step_name, now)
                         
                         if (count_val or 0) >= limit:
                             continue

                    # Claim it
                    # We already locked it, so update is safe
                    updated_row = await conn.fetchrow("""
                        UPDATE tasks
                        SET state='running', worker_id=$1, lease_expires_at=$2, started_at=$3
                        WHERE id=$4
                        RETURNING *
                    """, worker_id, expires_at, now, row["id"])
                    
                    if updated_row:
                        return self._row_to_task(updated_row)
        return None

    async def renew_task_lease(
        self,
        task_id: str,
        worker_id: str,
        lease_duration: timedelta,
    ) -> bool:
        assert self.pool is not None
        expires_at = datetime.now() + lease_duration
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow(
                """
                UPDATE tasks
                SET lease_expires_at=$1
                WHERE id=$2 AND worker_id=$3 AND state='running'
                RETURNING id
                """,
                expires_at,
                task_id,
                worker_id,
            )
            return row is not None

    async def list_tasks_for_execution(self, execution_id: str) -> List[TaskRecord]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = await conn.fetch("SELECT * FROM tasks WHERE execution_id = $1", execution_id)
            return [self._row_to_task(row) for row in rows]

    async def append_progress(self, execution_id: str, progress: ExecutionProgress) -> None:
         assert self.pool is not None
         async with self.pool.acquire() as conn:
             await conn.execute(
                "INSERT INTO execution_progress (execution_id, step, status, started_at, completed_at, detail) VALUES ($1, $2, $3, $4, $5, $6)",
                execution_id, progress.step, progress.status, progress.started_at, progress.completed_at, progress.detail
            )

    async def get_cached_result(self, cache_key: str) -> bytes | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow("SELECT value, expires_at FROM cache WHERE key = $1", cache_key)
            if row:
                val, expires_at = row['value'], row['expires_at']
                # logger.debug(f"Fetched from cache: key={cache_key}, expires_at={expires_at}, value_len={len(val) if val else 0}")
                if expires_at and expires_at < datetime.now():
                    # logger.debug(f"Cache expired for key={cache_key}")
                    return None
                return val
            # logger.debug(f"Cache miss for key={cache_key}")
        return None

    async def set_cached_result(self, cache_key: str, value: bytes, ttl: timedelta | None = None) -> None:
        expires_at = None
        if ttl:
            expires_at = datetime.now() + ttl
        
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.execute("""
                INSERT INTO cache (key, value, expires_at) VALUES ($1, $2, $3)
                ON CONFLICT (key) DO UPDATE SET value=$2, expires_at=$3
            """, cache_key, value, expires_at)

    async def get_idempotency_result(self, idempotency_key: str) -> bytes | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow("SELECT value FROM idempotency WHERE key = $1", idempotency_key)
            return row['value'] if row else None

    async def set_idempotency_result(self, idempotency_key: str, value: bytes) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
             await conn.execute("""
                INSERT INTO idempotency (key, value) VALUES ($1, $2)
                ON CONFLICT (key) DO UPDATE SET value=$2
            """, idempotency_key, value)

    async def move_task_to_dead_letter(self, task: TaskRecord, reason: str) -> None:
         assert self.pool is not None
         async with self.pool.acquire() as conn:
            payload = task_record_to_json(task)
            await conn.execute(
                "INSERT INTO dead_tasks (id, reason, moved_at, data) VALUES ($1, $2, $3, $4)",
                task.id, reason, datetime.now(), payload
            )

    def _row_to_dead_letter(self, row: Any) -> DeadLetterRecord:
        return row_to_dead_letter(row)

    async def list_dead_tasks(self, limit: int = 50) -> List[DeadLetterRecord]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = await conn.fetch(
                "SELECT * FROM dead_tasks ORDER BY moved_at DESC LIMIT $1",
                limit,
            )
            return [self._row_to_dead_letter(row) for row in rows]

    async def count_dead_tasks(self) -> int:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            val = await conn.fetchval("SELECT COUNT(*) FROM dead_tasks")
            return val if val else 0

    async def get_dead_task(self, task_id: str) -> DeadLetterRecord | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow(
                "SELECT * FROM dead_tasks WHERE id = $1",
                task_id,
            )
            if not row:
                return None
            return self._row_to_dead_letter(row)

    async def delete_dead_task(self, task_id: str) -> bool:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            tag = await conn.execute("DELETE FROM dead_tasks WHERE id = $1", task_id)
            try:
                count = int(tag.split(" ")[1])
            except (IndexError, ValueError):
                count = 0
            return count > 0

    async def cleanup_executions(self, older_than: datetime) -> int:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            where_clause = "completed_at IS NOT NULL AND completed_at < $1 AND state IN ('completed', 'failed', 'timed_out', 'cancelled')"
            
            async with conn.transaction():
                # Delete dependents using subquery
                await conn.execute(f"DELETE FROM tasks WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", older_than)
                await conn.execute(f"DELETE FROM execution_progress WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", older_than)
                await conn.execute(f"DELETE FROM signals WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", older_than)
                await conn.execute(f"DELETE FROM execution_counters WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", older_than)
                await conn.execute(f"DELETE FROM execution_state WHERE execution_id IN (SELECT id FROM executions WHERE {where_clause})", older_than)
                
                # Delete executions
                tag = await conn.execute(f"DELETE FROM executions WHERE {where_clause}", older_than)
                # tag format is usually "DELETE <count>"
                try:
                    count = int(tag.split(" ")[1])
                except (IndexError, ValueError):
                    count = 0
                return count

    async def cleanup_dead_letters(self, older_than: datetime) -> int:
        """Remove dead letter records older than the specified datetime."""
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            tag = await conn.execute(
                "DELETE FROM dead_tasks WHERE moved_at < $1",
                older_than,
            )
            try:
                count = int(tag.split(" ")[1])
            except (IndexError, ValueError):
                count = 0
            return count

    async def create_signal(self, signal: SignalRecord) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.execute("""
                INSERT INTO signals (execution_id, name, payload, created_at, consumed, consumed_at) VALUES ($1, $2, $3, $4, $5, $6)
                ON CONFLICT (execution_id, name) DO UPDATE SET payload=$3, consumed=$5, consumed_at=$6
            """, signal.execution_id, signal.name, signal.payload, signal.created_at, signal.consumed, signal.consumed_at)

    async def get_signal(self, execution_id: str, name: str) -> SignalRecord | None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            row = await conn.fetchrow("SELECT * FROM signals WHERE execution_id = $1 AND name = $2", execution_id, name)
            if not row:
                return None
            return row_to_signal(row)

    async def ensure_execution_counters(self, execution_id: str, counters: dict[str, int | float]) -> None:
        if not counters:
            return
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = [(execution_id, name, float(value)) for name, value in counters.items()]
            await conn.executemany(
                """
                INSERT INTO execution_counters (execution_id, name, value)
                VALUES ($1, $2, $3)
                ON CONFLICT (execution_id, name) DO NOTHING
                """,
                rows,
            )

    async def increment_execution_counter(self, execution_id: str, name: str, amount: int | float) -> float:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            value = await conn.fetchval(
                """
                INSERT INTO execution_counters (execution_id, name, value)
                VALUES ($1, $2, $3)
                ON CONFLICT (execution_id, name) DO UPDATE SET value = execution_counters.value + EXCLUDED.value
                RETURNING value
                """,
                execution_id,
                name,
                float(amount),
            )
            return float(value or 0.0)

    async def get_execution_counters(self, execution_id: str) -> dict[str, int | float]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = await conn.fetch(
                "SELECT name, value FROM execution_counters WHERE execution_id = $1",
                execution_id,
            )
            return {row["name"]: row["value"] for row in rows}

    async def ensure_execution_state_values(self, execution_id: str, values: dict[str, bytes]) -> None:
        if not values:
            return
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = [(execution_id, key, value) for key, value in values.items()]
            await conn.executemany(
                """
                INSERT INTO execution_state (execution_id, key, value)
                VALUES ($1, $2, $3)
                ON CONFLICT (execution_id, key) DO NOTHING
                """,
                rows,
            )

    async def set_execution_state_value(self, execution_id: str, key: str, value: bytes) -> None:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            await conn.execute(
                """
                INSERT INTO execution_state (execution_id, key, value)
                VALUES ($1, $2, $3)
                ON CONFLICT (execution_id, key) DO UPDATE SET value = EXCLUDED.value
                """,
                execution_id,
                key,
                value,
            )

    async def get_execution_state_values(self, execution_id: str) -> dict[str, bytes]:
        assert self.pool is not None
        async with self.pool.acquire() as conn:
            rows = await conn.fetch(
                "SELECT key, value FROM execution_state WHERE execution_id = $1",
                execution_id,
            )
            return {row["key"]: row["value"] for row in rows}


    def _policy_to_json(self, p: RetryPolicy | None) -> str:
        return retry_policy_to_json(p)

    def _json_to_policy(self, s: str) -> RetryPolicy:
        return retry_policy_from_json(s)

    def _row_to_execution(self, row: Any, progress: List[ExecutionProgress]) -> ExecutionRecord:
        return row_to_execution(row, progress)

    def _row_to_progress(self, row: Any) -> ExecutionProgress:
        return row_to_progress(row)

    def _row_to_task(self, row: Any) -> TaskRecord:
        return row_to_task(row)



import asyncio
import os
import pytest
from stent import Stent

def get_test_backend(test_id: str):
    backend_type = os.environ.get("STENT_TEST_BACKEND", "sqlite")
    
    if backend_type == "postgres":
        # Assumes a running postgres instance, e.g. via docker
        # DSN format: postgres://user:password@host:port/database
        # We append test_id to db name or use schemas? 
        # For simplicity, let's assume a single test db and we clear tables or use unique prefixes if needed.
        # BUT, standard practice for concurrent tests is separate DBs or schemas.
        # Here we will rely on the caller to cleanup or we can use a single DB and init_db ensures tables exist.
        dsn = os.environ.get("STENT_TEST_PG_DSN", "postgres://postgres:postgres@localhost:5432/stent_test")
        backend = Stent.backends.PostgresBackend(dsn)
        return backend
    else:
        db_path = f"test_stent_{test_id}.sqlite"
        return Stent.backends.SQLiteBackend(db_path)

async def clear_test_backend(backend):
    """Clears tables but keeps connection open if possible"""
    if hasattr(backend, "pool") and backend.pool:
         async with backend.pool.acquire() as conn:
             # Check if tables exist before truncating
              exists = await conn.fetchval(
                 "SELECT EXISTS (SELECT FROM information_schema.tables WHERE table_name = 'executions')"
              )
              if exists:
                 await conn.execute("TRUNCATE TABLE executions, execution_progress, tasks, dead_tasks, cache, idempotency, execution_counters, execution_state CASCADE")

async def cleanup_test_backend(backend):
    # First close the backend connection
    if hasattr(backend, "close"):
        await backend.close()
    
    if hasattr(backend, "db_path"):
        if not os.path.exists(backend.db_path):
            return

        for attempt in range(20):
            try:
                os.remove(backend.db_path)
                return
            except PermissionError:
                # Windows keeps SQLite files locked briefly after connections close.
                if attempt == 19:
                    return
                await asyncio.sleep(0.05 * (attempt + 1))
    elif hasattr(backend, "pool") and backend.pool:
         # For postgres, we might want to truncate tables
         async with backend.pool.acquire() as conn:
             await conn.execute("TRUNCATE TABLE executions, execution_progress, tasks, dead_tasks, cache, idempotency, execution_counters, execution_state CASCADE")
         await backend.pool.close()



import asyncio
import os
import logging
from stent import Stent, Result, RetryPolicy

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s: %(message)s', datefmt='%H:%M:%S')
logger = logging.getLogger("SagaExample")

# --- Activities (Steps) ---

@Stent.durable()
async def book_flight(trip_id: str) -> str:
    # Simulate success
    await asyncio.sleep(0.2)
    logger.info(f"Flight booked for {trip_id}")
    return f"flight_{trip_id}"

@Stent.durable()
async def cancel_flight(trip_id: str):
    await asyncio.sleep(0.2)
    logger.warning(f"Flight cancelled for {trip_id}")
    return True

@Stent.durable()
async def book_hotel(trip_id: str) -> str:
    await asyncio.sleep(0.2)
    logger.info(f"Hotel booked for {trip_id}")
    return f"hotel_{trip_id}"

@Stent.durable()
async def cancel_hotel(trip_id: str):
    await asyncio.sleep(0.2)
    logger.warning(f"Hotel cancelled for {trip_id}")
    return True

@Stent.durable()
async def book_car(trip_id: str) -> str:
    await asyncio.sleep(0.2)
    # Simulate failure for a specific trip ID to test compensation
    if "fail_car" in trip_id:
        raise ValueError("No cars available!")
    logger.info(f"Car booked for {trip_id}")
    return f"car_{trip_id}"

@Stent.durable()
async def cancel_car(trip_id: str):
    await asyncio.sleep(0.2)
    logger.warning(f"Car cancelled for {trip_id}")
    return True

# --- Orchestrator (Saga) ---

@Stent.durable()
async def trip_booking_saga(trip_id: str) -> Result[dict, Exception]:
    compensations = []
    data = {}
    
    try:
        # Step 1: Flight
        flight_ref = await book_flight(trip_id)
        data["flight"] = flight_ref
        compensations.append(lambda: cancel_flight(trip_id))
        
        # Step 2: Hotel
        hotel_ref = await book_hotel(trip_id)
        data["hotel"] = hotel_ref
        compensations.append(lambda: cancel_hotel(trip_id))
        
        # Step 3: Car (might fail)
        car_ref = await book_car(trip_id)
        data["car"] = car_ref
        # compensations.append(lambda: cancel_car(trip_id)) # Not needed if last step
        
        return Result.Ok(data)

    except Exception as e:
        logger.error(f"Saga failed: {e}. Starting compensation...")
        # Execute compensations in reverse order
        for comp in reversed(compensations):
            try:
                await comp()
            except Exception as ce:
                logger.error(f"Compensation failed: {ce}")
        
        return Result.Error(e)

# --- Runner ---

async def main():
    db_path = "saga_example.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
    
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    executor = Stent(backend=backend)
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.1))
    
    # 1. Successful Trip
    logger.info("--- Starting Successful Trip ---")
    exec_id_1 = await executor.dispatch(trip_booking_saga, "trip_hawaii")
    await _wait(executor, exec_id_1)
    
    # 2. Failed Trip (Car failure triggers rollbacks)
    logger.info("--- Starting Failed Trip (Saga Rollback) ---")
    exec_id_2 = await executor.dispatch(trip_booking_saga, "trip_fail_car")
    await _wait(executor, exec_id_2)
    
    worker_task.cancel()
    try:
        await worker_task
    except asyncio.CancelledError:
        pass

async def _wait(executor, exec_id):
    while True:
        state = await executor.state_of(exec_id)
        if state.state in ("completed", "failed", "timed_out"):
            break
        await asyncio.sleep(0.2)
    
    # The saga handles internal errors and returns Result.Error, so state might be 'completed'
    # but result is Error.
    res = await executor.result_of(exec_id)
    if res.ok:
        logger.info(f"Saga Result: SUCCESS - {res.value}")
    else:
        logger.info(f"Saga Result: FAILED (Handled) - {res.error}")

if __name__ == "__main__":
    asyncio.run(main())



from __future__ import annotations

import asyncio
from datetime import datetime, timedelta
from logging import Logger
from typing import TYPE_CHECKING, Any, Callable, TypeVar, cast

from stent.core import ExecutionProgress, ExecutionRecord, RetryPolicy, TaskRecord, compute_retry_delay
from stent.executor_context import ExecutionContext, current_execution_context
from stent.utils.idempotency import default_idempotency_key

if TYPE_CHECKING:
    from stent.executor import Stent

TExpiryError = TypeVar("TExpiryError", bound=BaseException)


async def lease_heartbeat_loop(
    *,
    executor: Stent,
    task_id: str,
    worker_id: str,
    lease_duration: timedelta,
    interval: timedelta,
    stop_event: asyncio.Event,
    logger: Logger,
) -> None:
    timeout = interval.total_seconds()
    while True:
        try:
            await asyncio.wait_for(stop_event.wait(), timeout=timeout)
            return
        except asyncio.TimeoutError:
            try:
                renewed = await executor.backend.renew_task_lease(task_id, worker_id, lease_duration)
            except Exception:
                logger.exception("Lease renewal failed for task %s", task_id)
                executor.metrics.lease_renewed(task_id=task_id, success=False)
                return

            executor.metrics.lease_renewed(task_id=task_id, success=renewed)
            if not renewed:
                logger.warning(
                    "Lease renewal lost for task %s on worker %s; allowing reclaim",
                    task_id,
                    worker_id,
                )
                return
        except asyncio.CancelledError:
            return


async def run_claimed_task(
    *,
    executor: Stent,
    task: TaskRecord,
    worker_id: str,
    lease_duration: timedelta,
    heartbeat_interval: timedelta | None,
    expiry_error_type: type[TExpiryError],
    unregistered_error_factory: Callable[[str], Exception],
    logger: Logger,
) -> None:
    execution = None
    heartbeat_task: asyncio.Task[Any] | None = None
    heartbeat_stop: asyncio.Event | None = None
    task_context = ExecutionContext(executor, task.execution_id)
    token_context = current_execution_context.set(task_context)
    try:
        execution = await executor.backend.get_execution(task.execution_id)
        if not execution:
            raise ValueError(f"Execution {task.execution_id} not found")

        if execution.expiry_at and datetime.now() > execution.expiry_at:
            execution.state = "timed_out"
            execution.completed_at = datetime.now()
            task.state = "failed"
            task.error = executor.serializer.dumps(Exception("Execution timed out"))
            task.completed_at = datetime.now()
            await executor.backend.update_execution(execution)
            await executor.backend.update_task(task)
            if executor.notification_backend:
                await executor.notification_backend.notify_task_updated(task.id, "failed")
                await executor.notification_backend.notify_execution_updated(task.execution_id, "timed_out")
            await executor.backend.append_progress(
                task.execution_id,
                ExecutionProgress(step=task.step_name, status="failed", detail="Execution timed out"),
            )
            return

        if execution.state in ("cancelling", "cancelled"):
            task.state = "failed"
            task.error = executor.serializer.dumps(Exception("Execution cancelled"))
            task.completed_at = datetime.now()
            await executor.backend.update_task(task)
            if executor.notification_backend:
                await executor.notification_backend.notify_task_updated(task.id, "failed")
            await executor.backend.append_progress(
                task.execution_id,
                ExecutionProgress(step=task.step_name, status="failed", detail="Execution cancelled"),
            )
            return

        if task.kind == "orchestrator" and execution.state == "pending":
            execution.state = "running"
            execution.started_at = datetime.now()
            await executor.backend.update_execution(execution)
            if executor.notification_backend:
                await executor.notification_backend.notify_execution_updated(task.execution_id, "running")

        args = executor.serializer.loads(task.args)
        kwargs = executor.serializer.loads(task.kwargs)

        meta = executor.registry.get(task.step_name)
        if not meta:
            raise unregistered_error_factory(task.step_name)

        await executor.backend.append_progress(
            task.execution_id,
            ExecutionProgress(step=task.step_name, status="running", started_at=datetime.now()),
        )

        if heartbeat_interval:
            heartbeat_stop = asyncio.Event()
            heartbeat_task = asyncio.create_task(
                executor._lease_heartbeat_loop(
                    task_id=task.id,
                    worker_id=worker_id,
                    lease_duration=lease_duration,
                    interval=heartbeat_interval,
                    stop_event=heartbeat_stop,
                )
            )

        if execution.expiry_at:
            remaining = (execution.expiry_at - datetime.now()).total_seconds()
            if remaining <= 0:
                raise expiry_error_type("Execution timed out before start")
            try:
                async with asyncio.timeout(remaining):
                    result_val = await meta.fn(*args, **kwargs)
            except asyncio.TimeoutError:
                execution.state = "timed_out"
                execution.completed_at = datetime.now()
                await executor.backend.update_execution(execution)

                task.state = "failed"
                task.error = executor.serializer.dumps(Exception("Execution timed out"))
                task.completed_at = datetime.now()
                await executor.backend.update_task(task)

                if executor.notification_backend:
                    await executor.notification_backend.notify_task_updated(task.id, "failed")
                    await executor.notification_backend.notify_execution_updated(task.execution_id, "timed_out")
                await executor.backend.append_progress(
                    task.execution_id,
                    ExecutionProgress(step=task.step_name, status="failed", detail="Execution timed out"),
                )
                return
        else:
            result_val = await meta.fn(*args, **kwargs)

        task.result = executor.serializer.dumps(result_val)
        task.state = "completed"
        task.completed_at = datetime.now()
        await executor.backend.update_task(task)

        await executor.backend.append_progress(
            task.execution_id,
            ExecutionProgress(
                step=task.step_name,
                status="completed",
                started_at=task.started_at,
                completed_at=datetime.now(),
            ),
        )

        if task.kind == "orchestrator":
            execution.result = task.result
            execution.state = "completed"
            execution.completed_at = datetime.now()
            await executor.backend.update_execution(execution)
            if executor.notification_backend:
                await executor.notification_backend.notify_execution_updated(task.execution_id, "completed")

        if executor.notification_backend:
            await executor.notification_backend.notify_task_completed(task.id)

        if task.kind == "orchestrator" and meta.cached:
            key = default_idempotency_key(meta.name, meta.version, args, kwargs, serializer=executor.serializer)
            await executor.backend.set_cached_result(key, task.result)
            logger.debug(f"Cached result for orchestrator {meta.name} with key {key}")

        duration_s = 0.0
        if task.started_at and task.completed_at:
            duration_s = (task.completed_at - task.started_at).total_seconds()
        executor.metrics.task_completed(
            queue=task.queue,
            step_name=task.step_name,
            kind=task.kind,
            duration_s=duration_s,
        )

    except Exception as e:
        retry_policy = task.retry_policy or RetryPolicy()
        attempt = task.retries + 1

        is_retryable = any(isinstance(e, t) for t in retry_policy.retry_for)

        if is_retryable and attempt < retry_policy.max_attempts:
            delay = compute_retry_delay(retry_policy, attempt)
            task.retries = attempt
            task.state = "pending"
            task.scheduled_for = datetime.now() + timedelta(seconds=delay)
            task.worker_id = None
            task.lease_expires_at = None
            task.started_at = None
            await executor.backend.update_task(task)
            await executor.backend.append_progress(
                task.execution_id,
                ExecutionProgress(step=task.step_name, status="failed", detail=str(e)),
            )
            executor.metrics.task_failed(
                queue=task.queue,
                step_name=task.step_name,
                kind=task.kind,
                reason=str(e),
                retrying=True,
            )
        else:
            task.state = "failed"
            task.error = executor.serializer.dumps(e)
            task.completed_at = datetime.now()
            await executor.backend.update_task(task)
            await executor.backend.move_task_to_dead_letter(task, str(e))
            await executor.backend.append_progress(
                task.execution_id,
                ExecutionProgress(step=task.step_name, status="failed", detail=str(e)),
            )
            executor.metrics.task_failed(
                queue=task.queue,
                step_name=task.step_name,
                kind=task.kind,
                reason=str(e),
                retrying=False,
            )
            executor.metrics.dead_lettered(
                queue=task.queue,
                step_name=task.step_name,
                kind=task.kind,
                reason=str(e),
            )
            if execution is None:
                raise ValueError(f"Execution {task.execution_id} not found")

            if task.kind == "orchestrator":
                execution_for_update = cast(ExecutionRecord, execution)
                execution_for_update.state = "failed"
                execution_for_update.error = task.error
                execution_for_update.completed_at = datetime.now()
                await executor.backend.update_execution(execution_for_update)
                if executor.notification_backend:
                    await executor.notification_backend.notify_execution_updated(task.execution_id, "failed")

            if executor.notification_backend:
                await executor.notification_backend.notify_task_updated(task.id, "failed")

    finally:
        try:
            await task_context.flush()
        finally:
            current_execution_context.reset(token_context)
        if heartbeat_stop:
            heartbeat_stop.set()
        if heartbeat_task:
            heartbeat_task.cancel()
            try:
                await heartbeat_task
            except asyncio.CancelledError:
                pass



import asyncio
import pytest
import os
from stent import Stent, Result

from datetime import timedelta

@Stent.durable()
async def signal_workflow():
    val = await Stent.wait_for_signal("my_signal")
    return val

@Stent.durable()
async def multi_signal_workflow():
    s1 = await Stent.wait_for_signal("sig1")
    s2 = await Stent.wait_for_signal("sig2")
    return s1 + s2

@pytest.mark.asyncio
async def test_signal_basic():
    db_path = "test_signals.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    executor = Stent(backend=backend)
    
    # Start worker
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.1))
    
    try:
        # Dispatch
        exec_id = await executor.dispatch(signal_workflow)
        
        # Wait a bit for it to reach wait state
        await asyncio.sleep(0.5)
        
        state = await executor.state_of(exec_id)
        # It might be running (waiting) or failed if bug
        if state.state == "failed":
            # Print error
            res = await executor.result_of(exec_id)
            print(res)
            
        assert state.state == "running"
        
        # Send signal
        await executor.send_signal(exec_id, "my_signal", "hello")
        
        # Wait for completion
        res = await executor.wait_for(exec_id, expiry=5.0)
        assert res.ok
        assert res.value == "hello"
        
    finally:
        worker_task.cancel()
        try:
            await worker_task
        except asyncio.CancelledError:
            pass
        await backend.close()
        if os.path.exists(db_path):
            os.remove(db_path)

@pytest.mark.asyncio
async def test_signal_buffered():
    """Test sending signal BEFORE workflow starts"""
    db_path = "test_signals_buffered.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    executor = Stent(backend=backend)
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.1))
    
    try:
        # Dispatch with delay
        exec_id = await executor.dispatch(signal_workflow, delay=timedelta(seconds=0.5)) 
        
        # Send signal immediately
        await executor.send_signal(exec_id, "my_signal", "early_bird")
        
        # Wait
        res = await executor.wait_for(exec_id, expiry=5.0)
        assert res.ok
        assert res.value == "early_bird"
        
    finally:
        worker_task.cancel()
        try:
            await worker_task
        except: pass
        await backend.close()
        if os.path.exists(db_path):
            os.remove(db_path)

@pytest.mark.asyncio
async def test_multi_signals():
    db_path = "test_multi_signals.sqlite"
    if os.path.exists(db_path):
        os.remove(db_path)
        
    backend = Stent.backends.SQLiteBackend(db_path)
    await backend.init_db()
    executor = Stent(backend=backend)
    worker_task = asyncio.create_task(executor.serve(poll_interval=0.1))
    
    try:
        exec_id = await executor.dispatch(multi_signal_workflow)
        
        await asyncio.sleep(0.5)
        await executor.send_signal(exec_id, "sig1", 10)
        
        await asyncio.sleep(0.5)
        await executor.send_signal(exec_id, "sig2", 20)
        
        res = await executor.wait_for(exec_id, expiry=5.0)
        assert res.ok
        assert res.value == 30
        
    finally:
        worker_task.cancel()
        try:
            await worker_task
        except: pass
        await backend.close()
        if os.path.exists(db_path):
            os.remove(db_path)



import sqlite3
import asyncio
from typing import Any, List, Optional

class AsyncCursor:
    def __init__(self, cursor: sqlite3.Cursor):
        self._cursor = cursor

    @property
    def rowcount(self):
        return self._cursor.rowcount

    async def fetchone(self) -> Any:
        return await asyncio.to_thread(self._cursor.fetchone)

    async def fetchall(self) -> List[Any]:
        return await asyncio.to_thread(self._cursor.fetchall)
    
    async def __aenter__(self):
        return self
        
    async def __aexit__(self, exc_type, exc, tb):
        pass 

class ExecuteContext:
    def __init__(self, conn, sql, parameters):
        self.conn = conn
        self.sql = sql
        self.parameters = parameters
        self.cursor = None

    def __await__(self):
        return self._execute().__await__()

    async def _execute(self):
        self.cursor = await self.conn._execute_impl(self.sql, self.parameters)
        return self.cursor

    async def __aenter__(self):
        self.cursor = await self._execute()
        return self.cursor

    async def __aexit__(self, exc_type, exc, tb):
        pass

class AsyncConnection:
    def __init__(self, path: str):
        self.path = path
        self._conn: Optional[sqlite3.Connection] = None

    async def __aenter__(self):
        self._conn = await asyncio.to_thread(sqlite3.connect, self.path, check_same_thread=False, detect_types=sqlite3.PARSE_DECLTYPES)
        return self

    async def __aexit__(self, exc_type, exc, tb):
        if self._conn:
            await asyncio.to_thread(self._conn.close)
            self._conn = None

    def execute(self, sql: str, parameters: tuple = ()) -> ExecuteContext:
        return ExecuteContext(self, sql, parameters)
        
    async def _execute_impl(self, sql, parameters):
        assert self._conn is not None
        cursor = await asyncio.to_thread(self._conn.execute, sql, parameters)
        return AsyncCursor(cursor)

    async def commit(self):
        assert self._conn is not None
        await asyncio.to_thread(self._conn.commit)
        
    @property
    def row_factory(self):
        assert self._conn is not None
        return self._conn.row_factory
    
    @row_factory.setter
    def row_factory(self, val):
        assert self._conn is not None
        self._conn.row_factory = val

def connect(path: str) -> AsyncConnection:
    return AsyncConnection(path)

def register_adapter(type, adapter):
    sqlite3.register_adapter(type, adapter)

def register_converter(typename, converter):
    sqlite3.register_converter(typename, converter)

Row = sqlite3.Row


import unittest
import asyncio
import os
import logging
import time
from stent import Stent, Result
from tests.utils import get_test_backend, cleanup_test_backend, clear_test_backend

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)

@Stent.durable()
async def parallel_worker(duration: float) -> float:
    await asyncio.sleep(duration)
    return duration

@Stent.durable()
async def parallel_orchestrator(count: int, duration: float) -> list[float]:
    tasks = []
    for _ in range(count):
        tasks.append(parallel_worker(duration))
    
    results = await asyncio.gather(*tasks)
    return results

@Stent.durable()
async def fan_out_fan_in_workflow(count: int, duration: float) -> float:
    results = await parallel_orchestrator(count, duration)
    return sum(results)

class TestParallel(unittest.IsolatedAsyncioTestCase):
    async def asyncSetUp(self):
        self.backend = get_test_backend(f"parallel_{os.getpid()}")
        await self.backend.init_db()
        await clear_test_backend(self.backend)
        self.executor = Stent(backend=self.backend)
        # We need high concurrency
        self.worker_task = asyncio.create_task(self.executor.serve(max_concurrency=10, poll_interval=0.1))

    async def asyncTearDown(self):
        self.worker_task.cancel()
        try:
            await self.worker_task
        except asyncio.CancelledError:
            pass
        await self.executor.shutdown()
        await cleanup_test_backend(self.backend)

    async def test_fan_out_fan_in(self):
        # We want to run 4 tasks, each sleeping 0.5s.
        # If sequential: 2.0s.
        # If parallel: ~0.5s + overhead.
        
        count = 4
        sleep_time = 0.5
        
        start_time = time.time()
        exec_id = await self.executor.dispatch(fan_out_fan_in_workflow, count, sleep_time)
        
        while True:
            state = await self.executor.state_of(exec_id)
            if state.state in ("completed", "failed", "timed_out"):
                break
            await asyncio.sleep(0.1)
            
        duration = time.time() - start_time
        result = await self.executor.result_of(exec_id)
        
        self.assertTrue(result.ok)
        self.assertEqual(result.value, count * sleep_time)
        
        # Check if it was actually parallel. 
        # Allow some overhead (e.g. 1.0s total instead of 0.5s is still better than 2.0s)
        # Overhead: DB polling (0.1s), task creation, etc.
        print(f"Total duration: {duration:.2f}s (expected < {count * sleep_time})")
        state = await self.executor.state_of(exec_id)
        # print(f"Execution state progress steps: {state.progress_str}")
        for progress in state.progress:
            print(f" - {progress.step}: {progress.status} (started at {progress.started_at}, completed at {progress.completed_at})")
        
        # Relaxed assertion for remote DB latency
        # We just want to ensure it's not strictly sequential (count * duration)
        # But with 4 tasks of 0.5s, sequential is 2.0s. 
        # If latency is huge (e.g. 14s total), we can't easily prove parallelism by time alone if overhead > task duration.
        # Instead, let's check the start/end times of the workers to see if they overlapped.
        
        worker_progress = [p for p in state.progress if p.step == "tests.test_parallel:parallel_worker" and p.status == "running"]
        # We need start/end times. But progress only has "running" start time or "completed" end time.
        # We need to correlate them or look at "completed" events which have start/end if available.
        # Actually ExecutionProgress struct has started_at and completed_at.
        
        completed_workers = [p for p in state.progress if p.step == "tests.test_parallel:parallel_worker" and p.status == "completed"]
        
        if len(completed_workers) == count:
            # Check overlap
            # Find the time window where most tasks were running
            intervals = [(p.started_at, p.completed_at) for p in completed_workers if p.started_at and p.completed_at]
            if intervals:
                # If any two intervals overlap, we have parallelism
                has_overlap = False
                for i in range(len(intervals)):
                    for j in range(i + 1, len(intervals)):
                        start_a, end_a = intervals[i]
                        start_b, end_b = intervals[j]
                        # Overlap if start_a < end_b and start_b < end_a
                        if start_a < end_b and start_b < end_a:
                            has_overlap = True
                            break
                    if has_overlap:
                        break
                
                self.assertTrue(has_overlap, "Tasks did not overlap, execution was sequential")
        else:
            # Fallback to duration check if we can't check timestamps
            self.assertLess(duration, count * sleep_time * 2.0, "Execution took too long")




import asyncio
from datetime import datetime, timedelta
from typing import List

import pytest

from stent.backend.sqlite import SQLiteBackend
from stent.core import ExecutionRecord, TaskRecord, RetryPolicy
from tests.utils import cleanup_test_backend


def _make_execution(exec_id: str) -> ExecutionRecord:
    now = datetime.now()
    return ExecutionRecord(
        id=exec_id,
        root_function="tests.durable",
        state="pending",
        args=b"{}",
        kwargs=b"{}",
        retries=0,
        created_at=now,
        started_at=None,
        completed_at=None,
        expiry_at=None,
        progress=[],
        tags=[],
        priority=0,
        queue=None,
    )


def _make_task(exec_id: str, task_id: str, step_name: str = "tests.step") -> TaskRecord:
    now = datetime.now()
    return TaskRecord(
        id=task_id,
        execution_id=exec_id,
        step_name=step_name,
        kind="activity",
        parent_task_id=None,
        state="pending",
        args=b"{}",
        kwargs=b"{}",
        retries=0,
        created_at=now,
        tags=[],
        priority=0,
        queue=None,
        retry_policy=RetryPolicy(),
    )


class FailingSQLiteBackend(SQLiteBackend):
    def __init__(self, db_path: str):
        super().__init__(db_path)
        self.fail_next_task_insert = False

    async def _insert_task(self, db, task: TaskRecord):  # type: ignore[override]
        if self.fail_next_task_insert:
            self.fail_next_task_insert = False
            raise RuntimeError("forced failure")
        await super()._insert_task(db, task)


@pytest.mark.asyncio
async def test_create_execution_with_root_task_is_atomic(tmp_path):
    backend = FailingSQLiteBackend(str(tmp_path / "atomic.sqlite"))
    await backend.init_db()

    record = _make_execution("exec-atomic")
    task = _make_task(record.id, "task-atomic")

    backend.fail_next_task_insert = True
    with pytest.raises(RuntimeError):
        await backend.create_execution_with_root_task(record, task)

    assert await backend.get_execution(record.id) is None
    assert await backend.get_task(task.id) is None

    await cleanup_test_backend(backend)


@pytest.mark.asyncio
async def test_sqlite_claim_next_task_multi_worker(tmp_path):
    backend = SQLiteBackend(str(tmp_path / "claim.sqlite"))
    await backend.init_db()

    exec_id = "exec-claim"
    await backend.create_execution(_make_execution(exec_id))

    tasks: List[TaskRecord] = [
        _make_task(exec_id, f"task-{idx}") for idx in range(50)
    ]
    await backend.create_tasks(tasks)

    async def worker(worker_idx: int):
        claimed: List[str] = []
        while True:
            task = await backend.claim_next_task(worker_id=f"worker-{worker_idx}")
            if not task:
                break
            claimed.append(task.id)
            task.state = "completed"
            task.completed_at = datetime.now()
            await backend.update_task(task)
        return claimed

    results = await asyncio.gather(*(worker(i) for i in range(6)))
    claimed_ids = [cid for worker_claims in results for cid in worker_claims]

    assert len(claimed_ids) == len(set(claimed_ids))
    assert set(claimed_ids) == {task.id for task in tasks}

    await cleanup_test_backend(backend)


@pytest.mark.asyncio
async def test_sqlite_concurrency_limits_respected(tmp_path):
    backend = SQLiteBackend(str(tmp_path / "limits.sqlite"))
    await backend.init_db()

    exec_id = "exec-limit"
    await backend.create_execution(_make_execution(exec_id))
    task_one = _make_task(exec_id, "task-one", step_name="only.step")
    task_two = _make_task(exec_id, "task-two", step_name="only.step")
    await backend.create_tasks([task_one, task_two])

    limits = {"only.step": 1}
    first_claim = await backend.claim_next_task(worker_id="worker-a", concurrency_limits=limits)
    assert first_claim is not None

    # Second claim should be blocked until the running task completes
    blocked_claim = await backend.claim_next_task(worker_id="worker-b", concurrency_limits=limits)
    assert blocked_claim is None

    first_claim.state = "completed"
    first_claim.completed_at = datetime.now()
    await backend.update_task(first_claim)

    second_claim = await backend.claim_next_task(worker_id="worker-c", concurrency_limits=limits)
    assert second_claim is not None

    await cleanup_test_backend(backend)


@pytest.mark.asyncio
async def test_sqlite_claim_respects_scheduled_for(tmp_path):
    backend = SQLiteBackend(str(tmp_path / "scheduled.sqlite"))
    await backend.init_db()

    exec_id = "exec-scheduled"
    await backend.create_execution(_make_execution(exec_id))

    task = _make_task(exec_id, "task-scheduled")
    task.scheduled_for = datetime.now() + timedelta(seconds=0.4)
    await backend.create_task(task)

    before_due = await backend.claim_next_task(worker_id="worker-a")
    assert before_due is None

    await asyncio.sleep(0.45)

    after_due = await backend.claim_next_task(worker_id="worker-a")
    assert after_due is not None
    assert after_due.id == "task-scheduled"

    await cleanup_test_backend(backend)


@pytest.mark.asyncio
async def test_dead_letter_round_trip(tmp_path):
    backend = SQLiteBackend(str(tmp_path / "dlq.sqlite"))
    await backend.init_db()

    exec_id = "exec-dlq"
    await backend.create_execution(_make_execution(exec_id))
    task = _make_task(exec_id, "task-dlq")
    await backend.create_task(task)

    await backend.move_task_to_dead_letter(task, "boom")
    records = await backend.list_dead_tasks()
    assert len(records) == 1
    record = await backend.get_dead_task(task.id)
    assert record is not None
    assert record.task.execution_id == exec_id
    deleted = await backend.delete_dead_task(task.id)
    assert deleted
    assert await backend.list_dead_tasks() == []

    await cleanup_test_backend(backend)


@pytest.mark.asyncio
async def test_sqlite_count_apis(tmp_path):
    backend = SQLiteBackend(str(tmp_path / "count.sqlite"))
    await backend.init_db()

    pending_exec = _make_execution("exec-pending")
    completed_exec = _make_execution("exec-completed")
    completed_exec.state = "completed"
    completed_exec.completed_at = datetime.now()

    await backend.create_execution(pending_exec)
    await backend.create_execution(completed_exec)

    assert await backend.count_executions() == 2
    assert await backend.count_executions(state="pending") == 1
    assert await backend.count_executions(state="completed") == 1

    dead_task = _make_task("exec-pending", "dead-task")
    await backend.move_task_to_dead_letter(dead_task, "boom")

    assert await backend.count_dead_tasks() == 1

    await cleanup_test_backend(backend)
